#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      plot.py
#  brief:     plot functions
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import sys
import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from scipy import interpolate

##########################################################################
#
# command arguments routines
#
##########################################################################


def add_arguments_units(parser):
    '''
    This function allow to add postscript options to a parser object
    '''

    parser.add_argument("--UnitLength_in_cm",
                      action="store",
                      dest="UnitLength_in_cm",
                      type=float,
                      default=None,
                      help="UnitLength in cm")

    parser.add_argument("--UnitMass_in_g",
                      action="store",
                      dest="UnitMass_in_g",
                      type=float,
                      default=None,
                      help="UnitMass in g")

    parser.add_argument("--UnitVelocity_in_cm_per_s",
                      action="store",
                      dest="UnitVelocity_in_cm_per_s",
                      type=float,
                      default=None,
                      help="UnitVelocity in cm per s")

    parser.add_argument("--param",
                      action="store",
                      dest="GadgetParameterFile",
                      type=str,
                      default=None,
                      help="Gadget parameter file",
                      metavar=" FILE")

    return parser




def apply_arguments_units(options):
    '''
    This function allow to extract units options from option object
    '''

    try:
        UnitLength_in_cm = options.UnitLength_in_cm
    except BaseException:
        UnitLength_in_cm = None

    try:
        UnitMass_in_g = options.UnitMass_in_g
    except BaseException:
        UnitMass_in_g = None

    try:
        UnitVelocity_in_cm_per_s = options.UnitVelocity_in_cm_per_s
    except BaseException:
        UnitVelocity_in_cm_per_s = None

    try:
        GadgetParameterFile = options.GadgetParameterFile
    except BaseException:
        GadgetParameterFile = None

    if GadgetParameterFile is not None:
        params = io.read_params(GadgetParameterFile)
    else:
        if UnitLength_in_cm is None or UnitMass_in_g is None or UnitVelocity_in_cm_per_s is None:
            params = None
        else:
            params = {}
            params['UnitLength_in_cm'] = UnitLength_in_cm
            params['UnitMass_in_g'] = UnitMass_in_g
            params['UnitVelocity_in_cm_per_s'] = UnitVelocity_in_cm_per_s

    return params


def add_arguments_reduc(parser):
    '''
    This function allow to reduc the number of particles
    '''

    parser.add_argument("--reduc",
                      action="store",
                      dest="reduc",
                      type=int,
                      default=None,
                      help="reduc from a factor n")

    return parser
    
    

def apply_arguments_reduc(nb, opt, verbose=True):

    if not hasattr(opt, "reduc"):
        print("Skipping reduc options..")
        return nb

    if isinstance(opt.reduc, int):
        if verbose:
            print(("reducing %s" % (opt.reduc)))
        nb = nb.reduc(opt.reduc)

    return nb



def add_arguments_center(parser):
    '''
    This function allow to center the model
    '''

    parser.add_argument(
        "--center",
        action="store",
        dest="center",
        type=str,
        default=None,
        help="center the model (histocenter,hdcenter,cmcenter,cmhistocenter)")

    parser.add_argument(
        "--inertial",
        action="store_true",
        dest="inertial",
        default=False,
        help="Switch to an inertial reference frame (not the case if PERIODICOUTER was used)")

    return parser
    
    

def apply_arguments_center(nb, opt, verbose=True):

    if not hasattr(opt, "center"):
        print("Skipping center options..")
        return nb

    # center the model
    if opt.center == 'hdcenter':
        if verbose:
            print(("centering %s" % (opt.center)))
        nb.hdcenter()

    elif opt.center == 'histocenter':
        if verbose:
            print(("centering %s" % (opt.center)))
        nb.histocenter()

    elif opt.center == 'cmcenter':
        if verbose:
            print(("centering %s" % (opt.center)))
        nb.cmcenter()

    elif opt.center == 'cmhistocenter':
        if verbose:
            print(("centering %s" % (opt.center)))
        nb.cmcenter()
        nb.histocenter()

    elif opt.center == 'rebox':
        if verbose:
            print(("centering %s" % (opt.center)))
        nb.rebox(mode='centred')

    if opt.inertial:
        print("Removing referential rotation")
        trace = nb.selectp(nb.trace_ids)
        tpos = np.mean(trace.pos, axis=0)
        pos = nb.pos - tpos
        nb.vel += np.cross(nb.trace_angular[:3], pos)
    return nb




def add_arguments_select(parser):
    '''
    This function allow to select particles from the model
    '''

    parser.add_argument(
        "--select",
        action="store",
        dest="select",
        type=str,
        default=None,
        help="select particles from the model ('gas','sph','sticky',...)")

    parser.add_argument("--ids_file",
                      action="store",
                      dest="ids_file",
                      type=str,
                      default=None,
                      help="File containing the requested IDs",
                      metavar="STR")

    parser.add_argument("--radius",
                      action="store",
                      dest="radius",
                      type=float,
                      default=None,
                      help="Max selection radius",
                      metavar="FLOAT")

    return parser



def apply_arguments_select(nb, opt, verbose=True):

    if not hasattr(opt, "select"):
        print("Skipping select options..")
        return nb

    # select
    if opt.select is not None:
        if verbose:
            print(("select %s" % (opt.select)))
        nb = nb.select(opt.select)

    if opt.ids_file is not None:
        if verbose:
            print(("selecting particles from %f" % opt.ids_file))
        nb = nb.selectp(file=opt.ids_file)

    if opt.radius is not None:
        if verbose:
            print(("removing particles at radius > %g" % opt.radius))
        nb = nb.selectc(nb.rxzy() < opt.radius)

    return nb


def add_arguments_info(parser):
    '''
    This function allow to select particles from the model
    '''

    parser.add_argument("--info",
                      action="store_true",
                      dest="info",
                      default=False,
                      help="print info on the model")

    return parser


def apply_arguments_info(nb, opt, verbose=True):

    if not hasattr(opt, "info"):
        print("Skipping info options..")
        return nb

    # select
    if opt.info:
        nb.info()

    return nb

def add_arguments_cmd(parser):
    '''
    This function allow to execute a command on the model
    '''

    parser.add_argument("--cmd",
                      action="store",
                      dest="cmd",
                      type=str,
                      default=None,
                      help="python command 'nb = nb.selectc((nb.T()>10))'")

    return parser


def apply_arguments_cmd(nb, opt, verbose=True):

    if not hasattr(opt, "cmd"):
        print("Skipping cmd options..")
        return nb

    # cmd
    if (opt.cmd is not None) and (opt.cmd != 'None'):
        if verbose:
            print(("exec : %s" % opt.cmd))
        
        # note : this is the new way of doing stuff in python3 :-(    
        namespace = {"nb":nb}
        exec(opt.cmd,namespace)
        nb = namespace["nb"]    
            
    return nb


def add_arguments_display(parser):
    '''
    This function allow to display the model
    '''

    parser.add_option("--display",
                      action="store",
                      dest="display",
                      type=str,
                      default=None,
                      help="display the model")

    return parser


def apply_arguments_display(nb, opt, verbose=True):

    if not hasattr(opt, "display"):
        print("Skipping display options..")
        return nb

    # display
    if (opt.display is not None) and (opt.display != 'None'):
        nb.display(
            obs=None,
            view=opt.display,
            marker='cross',
            pt=None,
            xp=None)

    return nb





def add_files_options(parser):
    '''
    This function allow to add color options to a parser object
    '''

    parser.add_argument("-t", "--ftype",
                      action="store",
                      dest="ftype",
                      type=str,
                      default='gh5',
                      help="type of the file",
                      metavar=" TYPE")

    parser.add_argument("--skip_io_block",
                      action="store",
                      dest="skip_io_block",
                      type=str,
                      default="",
                      help="Skip IO block (comma separated list)",
                      metavar="STR")

    parser.add_argument("--verbose",
                      action="store",
                      dest="verbose",
                      type=int,
                      default=0,
                      help="Define the verbose level (0: minimal, 1: standard,"
                      " 2: details 10:maximal)",
                      metavar="INT")

    return parser



def add_comoving_options(parser):
    """
    This function allow to force comoving integration for a snapshot.
    """

    parser.add_argument("--forceComovingIntegrationOn",
                      action="store_true",
                      dest="forceComovingIntegrationOn",
                      default=False,
                      help="force the model to be in in comoving integration")

    parser.add_argument(
        "--forceComovingIntegrationOff",
        action="store_true",
        dest="forceComovingIntegrationOff",
        default=False,
        help="force the model not to be in in comoving integration")

    return parser


def apply_arguments_comoving(nb, opt):

    if not hasattr(opt, "forceComovingIntegrationOn"):
        print("Skipping comoving options..")
        return nb

    if opt.forceComovingIntegrationOn:
        nb.setComovingIntegrationOn()

    if opt.forceComovingIntegrationOff:
        nb.setComovingIntegrationOff()

    return nb



def add_arguments_legend(parser):
    '''
    This function allow to add legend options to a parser object
    '''

    parser.add_argument("--legend",
                      action="store_true",
                      dest="legend",
                      default=False,
                      help="add a legend")

    parser.add_argument("--legend_ncol",
                      action="store",
                      dest="legend_ncol",
                      default=1,
                      type=int,
                      help="number of columns for the legend")

    parser.add_argument("--legend_txt",
                      action="store",
                      dest="legend_txt",
                      type=str,
                      default=None,
                      help="legend text",
                      metavar=" LIST of STRINGS")

    parser.add_argument("--legend_loc",
                      action="store",
                      type=str,
                      dest="legend_loc",
                      default=None,
                      help="legend location 'upper right'... ")

    return parser



def add_arguments_icshift(parser):
    '''
    This function allow to shift the model if needed
    '''

    parser.add_argument("--remove_ic_shift",
                      action="store_true",
                      dest="remove_ic_shift",
                      default=False,
                      help="remove the shift applied on the initial conditions.")

    return parser

def apply_arguments_icshift(nb, opt):

    if nb.has_var("InitialConditions_shift"):
      if opt.remove_ic_shift:
        nb.message("translate back : %s"%nb.InitialConditions_shift)
        nb.translate(-nb.InitialConditions_shift)
    
    return nb


##########################################################################
#
# graph limits routines
#
##########################################################################


def SetLimitsFromDataPoints(datas, xmin, xmax, ymin, ymax, log=None):

  if len(datas) == 1:
      x = datas[0].x
      y = datas[0].y
  else:
      x = np.array([], float)
      y = np.array([], float)
  
      for data in datas:
          x = np.concatenate((x, data.x))
          y = np.concatenate((y, data.y))
  
  return SetLimits(xmin, xmax, ymin, ymax, x, y, log=log)


def SetLimits(xmin, xmax, ymin, ymax, x, y, log=None):

    if log is not None:
        if str.find(log, 'x') != -1:

            x, y = CleanVectorsForLogX(x, y)

            x = np.log10(x)
            if xmin is not None:
                xmin = np.log10(xmin)
            if xmax is not None:
                xmax = np.log10(xmax)

    #############################
    # set x

    if xmin is None:
        xmin = min(x)

    if xmax is None:
        xmax = max(x)

    if xmin == xmax:
        xmin = xmin - 0.05 * xmin
        xmax = xmax + 0.05 * xmax
    else:
        xmin = xmin - 0.05 * (xmax - xmin)
        xmax = xmax + 0.05 * (xmax - xmin)

    # cut y values based on x
    #c = (x>=xmin)*(x<=xmax)
    #y = compress(c,y)

    if log is not None:
        if str.find(log, 'x') != -1:
            # if log=='x' or log=='xy':

            xmin = 10**xmin
            xmax = 10**xmax

    #############################
    # set y

    if log is not None:
        if str.find(log, 'y') != -1:

            x, y = CleanVectorsForLogY(x, y)

            y = np.log10(y)
            if ymin is not None:
                ymin = np.log10(ymin)
            if ymax is not None:
                ymax = np.log10(ymax)

    if ymin is None:
        ymin = min(y)

    if ymax is None:
        ymax = max(y)

    if ymin == ymax:
        ymin = ymin - 0.05 * ymin
        ymax = ymax + 0.05 * ymax
    else:
        ymin = ymin - 0.05 * (ymax - ymin)
        ymax = ymax + 0.05 * (ymax - ymin)

    if log is not None:
        if str.find(log, 'y') != -1:
            # if log=='y' or log=='xy':
            ymin = 10**ymin
            ymax = 10**ymax

    return xmin, xmax, ymin, ymax, log






def SetAxis(ax,xmin, xmax, ymin, ymax, log=None, extend=True):
  """
  Set ticks for the axis
  (this is the new routine)
  """
  
  #####################################
  # first : slightly extend the limits
  #####################################
  
  if extend:
      
    if log is not None:
        if str.find(log, 'x') != -1:
            if xmin is not None:
                xmin = np.log10(xmin)
            if xmax is not None:
                xmax = np.log10(xmax)
    
    if log is not None:
        if str.find(log, 'y') != -1:
            if ymin is not None:
                ymin = np.log10(ymin)
            if ymax is not None:
                ymax = np.log10(ymax)
                
    
    if xmin is not None and xmax is not None:
      if xmin == xmax:
          xmin = xmin - 0.05 * xmin
          xmax = xmax + 0.05 * xmax
      else:
          xmin = xmin - 0.05 * (xmax - xmin)
          xmax = xmax + 0.05 * (xmax - xmin)  
    
    if ymin is not None and ymax is not None:
      if ymin == ymax:
          ymin = ymin - 0.05 * ymin
          ymax = ymax + 0.05 * ymax
      else:
          ymin = ymin - 0.05 * (ymax - ymin)
          ymax = ymax + 0.05 * (ymax - ymin)
    
    
    if log is not None:
        if str.find(log, 'x') != -1:
            xmin = 10**xmin
            xmax = 10**xmax
    
    if log is not None:
        if str.find(log, 'y') != -1:
            ymin = 10**ymin
            ymax = 10**ymax


  #####################################
  # second : set log log or lin log
  #####################################


  if log is not None:
      if str.find(log, 'x') != -1 and str.find(log, 'y') != -1:
          ax.loglog()
      elif str.find(log, 'x') != -1:
          ax.semilogx()
      else:
          ax.semilogy()

  plt.axis([xmin, xmax, ymin, ymax])


  if log is None:
      log = 'None'

  #####################################
  # third : adapt ticks
  #####################################
  
  if str.find(log, 'x') == -1:
      ax.xaxis.set_major_locator(plt.AutoLocator())
      x_major = ax.xaxis.get_majorticklocs()
      dx_minor = (x_major[-1] - x_major[0]) / (len(x_major) - 1) / 5.
      ax.xaxis.set_minor_locator(plt.MultipleLocator(dx_minor))


  if str.find(log, 'y') == -1:
      ax.yaxis.set_major_locator(plt.AutoLocator())
      y_major = ax.yaxis.get_majorticklocs()
      dy_minor = (y_major[-1] - y_major[0]) / (len(y_major) - 1) / 5.
      ax.yaxis.set_minor_locator(plt.MultipleLocator(dy_minor))









##########################################################################
#
# Vector Cleaning
#
##########################################################################

def CleanVectors(x, y):
    '''
    remove bad values
    '''

    c = np.isfinite(x) * np.isfinite(y)
    x = np.compress(c, x)
    y = np.compress(c, y)

    return x.astype(float), y.astype(float)


def CleanVectorsForLogX(x, y):
    '''
    remove negative values
    '''

    c = (x > 0)
    x = np.compress(c, x)
    y = np.compress(c, y)

    return x.astype(float), y.astype(float)


def CleanVectorsForLogY(x, y):
    '''
    remove negative values
    '''

    c = (y > 0)
    x = np.compress(c, x)
    y = np.compress(c, y)

    return x.astype(float), y.astype(float)




##########################################################################
#
# The Colors class
#
##########################################################################


class ColorList():    
    '''
    Handle a list of colors
    '''

    def __init__(self, colormap=None, n=256, clist=None):
        '''
        Initialize with two modes :

        1) give number + eventually palette name
        2) give the list of colors
        '''

        if clist is not None:
            self.ls = clist
        else:
          if colormap is None:
            cmap = mpl.colormaps['jet']
          else:
            cmap = mpl.colormaps[colormap]
          
          self.ls = []
          
          # set colors
          ii = np.linspace(0,255,n).astype(int)
          for i in ii:
            self.ls.append(cmap(i))

        self.i = 0



    def set(self, i):
        try:
            self.i = self.ls.index(i)
        except BaseException:
            pass

    def current(self):
        return self.ls[self.i]

    def __next__(self):
        self.i = self.i + 1
        if self.i == len(self.ls):
            self.i = 0
        return self.ls[self.i]

    def get(self):

        ls = self.ls[self.i]

        self.i = self.i + 1
        if self.i == len(self.ls):
            self.i = 0

        return ls



##########################################################################
#
# Data point class
#
##########################################################################


class DataPoints():

    def __init__(
            self,
            x,
            y,
            z=None,
            yerr=None,
            color='k',
            linestyle='-',
            pointmarker='.',
            label='',
            tpe='line',
            linewidth=1):

        self.x = x
        self.y = y
        self.z = z
        self.yerr = yerr

        self.color = color
        self.linestyle = linestyle
        self.linewidth = linewidth
        self.pointmarker = pointmarker
        self.label = label

        self.tpe = tpe

    def reduc(self, rf, n=None, mn=None, mx=None, dx=None):

        if dx is not None:
            n = int((max(self.x) - min(self.x)) / dx)
            yh, xh = np.histogram(
                self.x, n, (min(
                    self.x), max(
                    self.x)), weights=self.y)
        else:
            if mn is None and mx is None and n is None:
                yh, xh = np.histogram(self.x, int(
                    len(self.x) / float(rf)), (min(self.x), max(self.x)), weights=self.y)
            else:
                yh, xh = np.histogram(self.x, n, (mn, mx), weights=self.y)

        xh = xh[:-1]

        self.x = xh
        self.y = yh

    def integrate(self):
        self.y = add.accumulate(self.y)

    def interpolate(self, xi):

        n = len(self.x)
        s1 = n - np.sqrt(2 * n)
        s2 = n + np.sqrt(2 * n)

        tck = interpolate.fitpack.splrep(self.x, self.y, s=s2, k=2)
        yi = interpolate.fitpack.splev(xi, tck)

        self.xi = xi
        self.yi = yi

    def derive(self):

        dx = self.x[1:] - self.x[0:-1]
        dy = self.y[1:] - self.y[0:-1]

        self.x = self.x[1:]
        self.y = dy / dx

        c = np.isfinite(self.y)

        self.x = np.compress(c, self.x)
        self.y = np.compress(c, self.y)

    def xy(self,):
        return self.x, self.y

    def get_x(self,):
        return self.x

    def get_y(self,):
        return self.y





def LegendFromDataPoints(ax,datas=None, loc=None, protect=True, opt=None):

    if opt is None:
        ncol = 1
    else:
        ncol = opt.legend_ncol

    if opt is not None and loc is None:
        if hasattr(opt, "legend_loc"):
            loc = opt.legend_loc

    if datas is not None:
        tags = []
        for data in datas:
            if protect:
                txt = data.label.replace("_", "\_")
            else:
                txt = data.label

            tags.append("%s" % txt)
        ax.legend(tags, loc=loc, ncol=ncol)
    else:
        ax.legend(loc=loc, ncol=ncol)







#################################
def getLOS(nlos,seed=None):
#################################
  """
  return n line of sights in for of an nx3 array
  """
  
  # get points in a shell of size 1
  if seed is not None:
    np.random.seed(seed=seed)
    
  rand1 = np.random.random(nlos)
  rand2 = np.random.random(nlos)
  
  # define a shell
  phi = rand1 * np.pi * 2.
  costh = 1. - 2. * rand2
  sinth = np.sqrt(1. - costh**2)
  
  x = sinth * np.cos(phi)
  y = sinth * np.sin(phi)
  z = costh
  
  los = np.transpose(np.array([x, y, z]))

  return los




#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      thermodyn.py
#  brief:     thermodynamics functions
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


import numpy as np
from . import ctes
from . import units
from . import iofunc as pnio

from pNbody import thermodynlib

##########################################################################
#
# THERMODYNAMIC RELATIONS
#
##########################################################################

# defaults parameters in cgs
defaultpars = {
    "k": ctes.BOLTZMANN,
    "mh": ctes.PROTONMASS,
    "mu": 2,
    "gamma": 5 / 3.,
    "G": ctes.GRAVITY}


# tabulation of the mean weight (from Grackle)
nt = 10
tt = np.array([1.0e+01, 1.0e+02, 1.0e+03, 1.0e+04, 1.3e+04,
            2.1e+04, 3.4e+04, 6.3e+04, 1.0e+05, 1.0e+09])
mt = np.array([1.18701555,
            1.15484424,
            1.09603514,
            0.9981496,
            0.96346395,
            0.65175895,
            0.6142901,
            0.6056833,
            0.5897776,
            0.58822635])


###################
def old_MeanWeightT(T):
    ###################
    """
    mean molecular weight as a function of the Temperature
    """

    if isinstance(T, np.ndarray):

        mu = np.zeros(len(T))
        for i in range(len(T)):
            mu[i] = MeanWeightT(T[i])

    else:

        logt = np.log(T)
        ttt = np.exp(logt)

        if ttt < tt[0]:
            j = 1
        else:
            for j in range(1, nt):
                if (ttt > tt[j - 1]) and (ttt <= tt[j]):
                    break

        slope = np.log(mt[j] / mt[j - 1]) / np.log(tt[j] / tt[j - 1])
        mu = np.exp(slope * (logt - np.log(tt[j])) + np.log(mt[j]))

    return mu


###################
def MeanWeightT(T):
    ###################
    """
    mean molecular weight as a function of the Temperature
    """

    return thermodynlib.MeanWeightT(T)


###################
def UNt(T):
    ###################
    """
    UN(T) = energy normalized as a function of T
          = T/mu(T)
    """

    return T / MeanWeightT(T)



# tabulation of the normalized energy vs T
unr = UNt(tt)


###################
def Tun(UN):
    ###################
    """
    T(UN) = temperature vs energy normalized

    inverse of UNt(U)

    """

    if isinstance(UN, np.ndarray):

        T = np.zeros(len(UN))
        for i in range(len(UN)):
            T[i] = Tun(UN[i])

    else:

        logu = np.log(UN)
        uuu = np.exp(logu)

        if uuu < unr[0]:
            j = 1
        else:
            for j in range(1, nt):
                if (uuu > unr[j - 1]) and (uuu <= unr[j]):
                    break

        slope = np.log(tt[j] / tt[j - 1]) / np.log(unr[j] / unr[j - 1])
        T = np.exp(slope * (logu - np.log(unr[j])) + np.log(tt[j]))

    return T


###################
def Prt(rho, T, pars=defaultpars):
    ###################
    """
    P(rho,T)
    """

    k = float(pars['k'])
    mh = float(pars['mh'])
    mu = float(pars['mu'])

    mumh = mu * mh

    return k * T / (mumh / rho)


###################
def Trp(rho, P, pars=defaultpars):
    ###################
    """
    T(rho,P)
    """

    k = float(pars['k'])
    mh = float(pars['mh'])
    mu = float(pars['mu'])

    mumh = mh * mu

    return (mumh / rho) / k * P


###################
def Art(rho, T, pars=defaultpars):
    ###################
    """
    A(rho,T)
    """

    k = float(pars['k'])
    mh = float(pars['mh'])
    mu = float(pars['mu'])
    gamma = float(pars['gamma'])

    mumh = mu * mh

    return k / mumh * rho**(1. - gamma) * T

###################
def Tra(rho, A, pars=defaultpars):
    ###################
    """
    T(rho,A)
    """

    k = float(pars['k'])
    mh = float(pars['mh'])
    mu = float(pars['mu'])
    gamma = float(pars['gamma'])

    mumh = mu * mh

    return mumh / k * rho**(gamma - 1.) * A


###################
def Urt(T, pars=defaultpars):
    ###################
    """
    U(rho,T)
    """

    k = float(pars['k'])
    mh = float(pars['mh'])
    gamma = float(pars['gamma'])

    U = UNt(T) / (gamma - 1.) * k / mh		# new, using the tabulated mu

    return U


###################
def Tru(U, pars=defaultpars):
    ###################
    """
    T(rho,U)
    """

    k = float(pars['k'])
    mh = float(pars['mh'])
    gamma = float(pars['gamma'])

    UN = (gamma - 1) * mh / k * U
    # T = Tun(UN)				# new, using the tabulated mu
    T = thermodynlib.Tun(UN)		# new, using the tabulated mu + C version

    return T


###################
def Tmuru(U, pars=defaultpars):
    ###################
    """
    T(rho,U)/mu  = UN
    """

    k = float(pars['k'])
    gamma = float(pars['gamma'])
    mh = float(pars['mh'])

    Tmu = (gamma - 1.) * mh / k * U

    return Tmu


###################
def Pra(rho, A, pars=defaultpars):
    ###################
    """
    P(rho,A)
    """

    gamma = float(pars['gamma'])

    return rho**gamma * A


###################
def Arp(rho, P, pars=defaultpars):
    ###################
    """
    A(rho,P)
    """

    gamma = float(pars['gamma'])

    return rho**-gamma * P


###################
def Pru(rho, U, pars=defaultpars):
    ###################
    """
    P(rho,U)
    """

    gamma = float(pars['gamma'])

    return (gamma - 1.) * rho * U


###################
def Urp(rho, P, pars=defaultpars):
    ###################
    """
    U(rho,P)
    """

    gamma = float(pars['gamma'])

    return 1. / (gamma - 1.) * (1 / rho) * P


###################
def Ura(rho, A, pars=defaultpars):
    ###################
    """
    U(rho,A)
    """

    gamma = float(pars['gamma'])

    return 1. / (gamma - 1.) * rho**(gamma - 1.) * A


###################
def Aru(rho, U, pars=defaultpars):
    ###################
    """
    A(rho,U)
    """

    gamma = float(pars['gamma'])

    return (gamma - 1.) * rho**(1. - gamma) * U


###################
def SoundSpeed_ru(U, pars=defaultpars):
    ###################
    """
    Sound Speed
    Cs(rho,U)
    """

    gamma = float(pars['gamma'])

    return np.sqrt((gamma - 1.) * gamma * U)


###################
def SoundSpeed_rt(T, pars=defaultpars):
    ###################
    """
    Sound Speed
    Cs(rho,T)
    """

    gamma = float(pars['gamma'])

    U = Urt(T, pars)

    return np.sqrt((gamma - 1.) * gamma * U)

###################
def JeansLength_ru(rho, U, pars=defaultpars):
    ###################
    """
    Jeans Length
    L_J(rho,U)
    """

    G = float(pars['G'])

    Cs = SoundSpeed_ru(U, pars)

    return Cs * np.sqrt(np.pi / (G * rho))


###################
def JeansLength_rt(rho, T, pars=defaultpars):
    ###################
    """
    Jeans Length
    L_J(rho,T)
    """

    G = float(pars['G'])

    Cs = SoundSpeed_rt(T, pars)

    return Cs * np.sqrt(np.pi / (G * rho))


###################
def JeansMass_ru(rho, U, pars=defaultpars):
    ###################
    """
    Jeans Mass
    M_J(rho,T)
    """

    G = float(pars['G'])

    Cs = SoundSpeed_ru(U, pars)

    return (np.pi**(5 / 2.) * Cs**3) / (6 * G**(3 / 2.) * np.sqrt(rho))


###################
def JeansMass_rt(rho, T, pars=defaultpars):
    ###################
    """
    Jeans Mass
    M_J(rho,T)
    """

    G = float(pars['G'])

    Cs = SoundSpeed_rt(T, pars)

    return (np.pi**(5 / 2.) * Cs**3) / (6 * G**(3 / 2.) * np.sqrt(rho))


###################
def MeanWeight(Xi, ionized=0):
    ###################
    """
    old version
    """

    if ionized:
        return 4. / (8. - 5. * (1. - Xi))
    else:
        return 4. / (1. + 3. * Xi)


###################
def ElectronDensity(rho, pars):
    ###################
    """
    Electron density for a mixture of H + He
    """

    Xi = float(pars['Xi'])
    mh = float(pars['mh'])
    ionisation = float(pars['ionisation'])

    if ionisation:
        return rho.astype(float) / mh * (Xi + (1 - Xi) / 2.)
    else:
        return rho.astype(float) * 0


#############################
def Lambda(rho, u, localsystem, thermopars, coolingfile):
    #############################
    """
    This corresponds to Lambda normalized

    Ln = L / nh 2

    nh = (xi*rho/mh)
    """

    UnitLength_in_cm = units.PhysCte(1, localsystem.UnitDic['m']).into(units.cgs)
    UnitTime_in_s = units.PhysCte(1, localsystem.UnitDic['s']).into(units.cgs)
    UnitMass_in_g = units.PhysCte(1, localsystem.UnitDic['kg']).into(units.cgs)

    UnitEnergy_in_cgs = UnitMass_in_g * UnitLength_in_cm**2 / UnitTime_in_s**2

    metalicity = thermopars['metalicity']
    hubbleparam = thermopars['hubbleparam']

    # compute cooling time
    logT, logL0, logL1, logL2, logL3, logL4, logL5, logL6 = pnio.read_cooling(
        coolingfile)

    if metalicity == 0:
        logL = logL0
    elif metalicity == 1:
        logL = logL1
    elif metalicity == 2:
        logL = logL2
    elif metalicity == 3:
        logL = logL3
    elif metalicity == 4:
        logL = logL4
    elif metalicity == 5:
        logL = logL5
    elif metalicity == 6:
        logL = logL6

    # compute gas temp
    logTm = np.log10(Tru(rho, u, thermopars))
    c = ((logTm >= 4) * (logTm < 8.5))
    u = np.where(c, u, 1)
    rho = np.where(c, rho, 1)
    # recompute gas temp
    logTm = np.log10(Tru(rho, u, thermopars))

    # get the right L for a given mT
    logLm = np.take(logL, np.searchsorted(logT, logTm))
    Lm = 10**logLm

    # transform in user units
    Lm = Lm / UnitEnergy_in_cgs / UnitLength_in_cm**3 * UnitTime_in_s

    L = Lm * hubbleparam

    return L


#############################
def CoolingTime(rho, u, localsystem, thermopars, coolingfile):
    #############################

    UnitLength_in_cm = units.PhysCte(1, localsystem.UnitDic['m']).into(units.cgs)
    UnitTime_in_s = units.PhysCte(1, localsystem.UnitDic['s']).into(units.cgs)
    UnitMass_in_g = units.PhysCte(1, localsystem.UnitDic['kg']).into(units.cgs)

    UnitEnergy_in_cgs = UnitMass_in_g * UnitLength_in_cm**2 / UnitTime_in_s**2

    ProtonMass = thermopars['mh']
    Xi = thermopars['Xi']
    metalicity = thermopars['metalicity']
    hubbleparam = thermopars['hubbleparam']

    # compute cooling time
    logT, logL0, logL1, logL2, logL3, logL4, logL5, logL6 = pnio.read_cooling(
        coolingfile)

    if metalicity == 0:
        logL = logL0
    elif metalicity == 1:
        logL = logL1
    elif metalicity == 2:
        logL = logL2
    elif metalicity == 3:
        logL = logL3
    elif metalicity == 4:
        logL = logL4
    elif metalicity == 5:
        logL = logL5
    elif metalicity == 6:
        logL = logL6

    # compute gas temp
    logTm = np.log10(Tru(rho, u, thermopars))
    c = ((logTm >= 4) * (logTm < 8.5))
    u = np.where(c, u, 1)
    rho = np.where(c, rho, 1)
    # recompute gas temp
    logTm = np.log10(Tru(rho, u, thermopars))

    # get the right L for a given mT
    logLm = np.take(logL, np.searchsorted(logT, logTm))
    Lm = 10**logLm

    # transform in user units
    Lm = Lm / UnitEnergy_in_cgs / UnitLength_in_cm**3 * UnitTime_in_s

    L = Lm * hubbleparam

    tc = (ProtonMass)**2 / (L * Xi**2) * u / rho
    tc = np.where(c, tc, 0)

    return tc, c



###########################################################################################
#  package:   Mockimgs
#  file:      detector.py
#  brief:     detector class
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
from astropy import units as u


class Detector():
  '''
  Define the Detector class.
  '''
  def __init__(self,name):    

    self.name       = name
    self.instrument = None

    # define the sensible
    self.defineSensibleZones()
    

  def info(self):
    """
    gives info on the detector
    """
    print("detector name              : %s"%self.name)

  def get_parameters(self,fmt=None):
    """
    return a dictionary containing usefull parameters    
    """
    params = {}
    params["detector_name"]           = self.detector
    return params  
  
  
  
  def getFocal(self):
    """
    if defined, get the instrument focal
    """
    if self.instrument is not None:
      return self.instrument.getFocal()
    else:
      raise ValueError("no instrument defined. Cannot get the telescope focal.")
    
  def getDistance(self):
    """
    if defined, get the object distance
    """
    if self.instrument is not None:
      obj = self.instrument.getObject()
      return obj.getDistance()
    else:
      raise ValueError("no object defined. Cannot get the object distance.")

  
  def toDistance(self,x):
    """
    convert distances in the ccd
    to physical distances, assuming some focal and distance
    """
    focal    = self.getFocal()
    distance = self.getDistance()
    return x/focal*distance
    
  
  def toAngle(self,x):  
    """
    get the factor to convert distances in the ccd
    to angle, assuming some focal    
    """
    focal    = self.getFocal()
    return np.arctan(x/focal)
      
  

  def defineSensibleZones(self):
    """
    set the sensible zone
    """
    xs = np.array([-1,1,1,-1])*u.mm
    ys = np.array([-1,-1,1,1])*u.mm    
    self.sZones = [Dzone(self,xs,ys)]
    
    
  def getSensibleZones(self):
    """
    return the sensible zones
    """  
    return self.sZones
  
  def draw(self,ax,mode=None,unit=None):
    """
    draw the detector
    """
    self.drawSensibleZones(ax,mode,unit)  
    
    
  def drawSensibleZones(self,ax,mode=None,unit=None):  
    """
    draw the sensible zones
  
    ax : matplotlib axis
    """
    
    focal    = self.getFocal()
    distance = self.getDistance()
    
    for sz in self.getSensibleZones():
      
      xs = sz.x(mode=mode,unit=unit,focal=focal,distance=distance)
      ys = sz.y(mode=mode,unit=unit,focal=focal,distance=distance)
      
      ax.scatter(xs,ys,c='k')
      
            
      
      
    
    
    
    


    
    
    



#!/usr/bin/env python3

###########################################################################################
#  package:   pNbody
#  file:      libRGBs.py
#  brief:     ???
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
###########################################################################################

import numpy as np
from pNbody import *

# Should be in its own class...
Kroupa_meanstar = 0.4820


def Kroupa_IMF(m):
        #Meanstar = 0.4820
    A = 0.10629
    if m <= 0.08:
        return(A * m**-0.3)
    elif m <= 0.5:
        return(A * m**-1.8)
    elif m <= 1:
        return(A * m**-2.7)
    else:
        return(A * m**-2.3)


def Kroupa_IMFint(m):
    A = 0.10629

    f = np.zeros(np.size(m))
    ind = np.where(m <= 0.08)
    f[ind] = A * m[ind]**0.7 / 0.7 - 0.01864976527083534

    ind = np.where((m > 0.08) * (m <= 0.5))
    f[ind] = -A * m[ind]**-.8 / .8 + 1.0021446918927897 + 0.0072656

    ind = np.where((m > 0.5) * (m <= 1))
    f[ind] = -A * m[ind]**-1.7 / 1.7 + 0.2031395463734218 + 0.778083245211

    ind = np.where(m > 1)
    f[ind] = -A * m[ind]**-1.3 / 1.3 + A * 1**-1.3 / 1.3 + 0.918699
    return(f)


class NRGB:

    def __init__(self, file):
        self.fileMS = file + 'MS.dat'
        self.fileRGB = file + 'rgb.dat'

        self.ReadMS()  # Read data file, create self.data
        self.ReadRGB()
        self.InterpolateMS()

        self.InterpolateTRGB()

    def ReadMS(self):
        pass

    def ReadRGB(self):
        pass

    def InterpolateMS(self):
        pass

    def InterpolateTRGB(self):
        pass

    Kroupa_meanstar = 0.4820

    def Kroupa_IMF(m):
        #Meanstar = 0.4820
        A = 0.10629
        if m <= 0.08:
            return(A * m**-0.3)
        elif m <= 0.5:
            return(A * m**-1.8)
        elif m <= 1:
            return(A * m**-2.7)
        else:
            return(A * m**-2.3)

    def Kroupa_IMFint(m):
        A = 0.10629

        f = np.zeros(np.size(m))
        f[np.where(m <= 0.08)] = A * m**0.7 / 0.7 - 0.01864976527083534
        f[np.where((m > 0.08) * (m <= 0.5))] = -A * \
            m**-.8 / .8 + 1.0021446918927897 + 0.0072656
        f[np.where((m > 0.5) * (m <= 1))] = -A * m**-1.7 / \
            1.7 + 0.2031395463734218 + 0.778083245211
        f[np.where(m > 1)] = -A * m**-1.3 / 1.3 + A * 1**-1.3 / 1.3 + 0.918699
        return(f)

    def RGBs(self, Zs, Ages):
        Frac = Kroupa_IMFint(self.InterpolateTRGB.ev(
            Zs, Ages)) - Kroupa_IMFint(self.InterpolateMS.ev(Zs, Ages))
        return(Frac / Kroupa_meanstar)



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      libdisk.py
#  brief:     libdisk.py
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

from pNbody import libgrid
import numpy as np


def get_Integral(v, dr, ia, ib):
    """
    Integrate the vector v, between ia and ib, using trapezes.


    v  : values of cells (must be 1 dimensional)
    dr : corresponding physical size of cells
    ia : lower  real indice			WARNING : the indicies must now be integer !!!
    ib : higher real indice			WARNING : the indicies must now be integer !!!
    """

    ia = max(0, ia)
    ib = min(len(v), ib)

    if ia == ib:
        return 0.0

    if ia > ib:
        raise Exception("ia must be < ib")


  # # below, the case of non int values for ia and ib
  #
  # iap = int(ceil(ia))
  # ibp = int(floor(ib))
  #
  # dra = iap-ia
  # drb = ib-ibp
  #
  # Ia = 0.0
  # if dra != 0:
  #   Ia = v[iap-1] * dra
  #
  # Ib = 0.0
  # if drb != 0:
  #   Ib = v[ibp] * drb
  #
  #
  # I = v[iap:ibp]*dr[iap:ibp]
  #
  # return sum(I)+Ia+Ib



    I = 0
    for i in range(ia, ib):
        I = I + dr[i] * (v[i + 1] + 0.5 * np.fabs(v[i + 1] - v[i]))

    return I


def Diff(f, x, s=None, k=2):
    """
    First derivative of f(x)
    """

    # if s!=None:
    #  tck = interpolate.fitpack.splrep(x,f,s=s,k=k)
    #  f  = interpolate.fitpack.splev(x,tck)

    fp = np.zeros(len(x), x.dtype)

    fp[0] = (f[1] - f[0]) / (x[1] - x[0])
    fp[-1] = (f[-1] - f[-2]) / (x[-1] - x[-2])

    f1 = f[2:]
    f2 = f[:-2]
    x1 = x[2:]
    x2 = x[:-2]

    fp[1:-1] = (f1 - f2) / (x1 - x2)

    return fp


def Kappa(R, dPhi, d2Phi):
  
    with np.errstate(divide='ignore'): # ignore warning that you divide by zero where R==0
        Kappa2 = np.where((R == 0), 0.0, d2Phi + dPhi * 3 / R)

    if R[0] == 0 and R[1] > 0:
        dPhi = 0.5 * (dPhi[0] + dPhi[1])
        d2Phi = 0.5 * (d2Phi[0] + d2Phi[1])
        R = 0.5 * R[1]
        Kappa2[0] = d2Phi + dPhi * 3 / R
    
    return np.sqrt(Kappa2)


def Omega(R, dPhi):

    with np.errstate(divide='ignore'): # ignore warning that you divide by zero where R==0
        Omega2 = np.where((R == 0), 0, dPhi / R)
    return np.sqrt(Omega2)


def Nu(z, Phi):

    nr, nz = Phi.shape

    odd = np.fmod(nz, 2)

    if odd:
        idx1 = (nz // 2)
        idx2 = (nz // 2) + 1
    else:
        idx1 = (nz // 2)

    nu2 = np.array([], np.float32)

    for i in range(nr):

        d1 = Diff(Phi[i, :], z)
        d2 = Diff(d1, z)

        if odd:
            nu2 = np.concatenate((nu2, np.array([0.5*(d2[idx1] + d2[idx2])])))
        else:
            nu2 = np.concatenate((nu2, np.array([d2[idx1]])))

    with np.errstate(divide='ignore'): # ignore warning that you divide by zero
        return np.where(nu2 > 0, np.sqrt(nu2), 0)


def Nu_Old(z, Phi):

    nr, nz = Phi.shape

    odd = np.fmod(nz, 2)

    if odd:
        idx1 = (nz - 1) / 2
    else:
        idx1 = (nz / 2) - 1
        idx2 = (nz / 2)

    nu2 = np.array([], np.float32)

    for i in range(nr):

        d1 = Diff(Phi[i, :], z)
        d2 = Diff(d1, z)

        if odd:
            nu2 = np.concatenate((nu2, np.array([d2[idx1]])))
        else:
            nu2 = np.concatenate((nu2, np.array([0.5 * (d2[idx1] + d2[idx2])])))

    return np.where(nu2 > 0, np.sqrt(nu2), 0)


def Vcirc2(R, dPhi):

    return R * dPhi


def Vcirc(R, dPhi):

    Vcirc2 = R * dPhi
    return np.sqrt(Vcirc2)


def QToomre(G, R, SigmaR, Kappa, Sdens):

    with np.errstate(divide='ignore'): # ignore warning that you divide by zero
        Q = np.where((Sdens == 0), 0, SigmaR * Kappa / (3.36 * G * Sdens))
    return Q


def XToomre(G, R, Kappa, Sdens, mode):
    with np.errstate(divide='ignore'): # ignore warning that you divide by zero
        X = np.where((Sdens == 0), 0, (R * Kappa**2) / (2 * np.pi * G * Sdens * mode))
    return X


def AAraki(SigmaR, Sigmaz):

    with np.errstate(divide='ignore'): # ignore warning that you divide by zero
        A = np.where((SigmaR == 0), 0, Sigmaz / SigmaR)
    return A


def get_1d_Sigma_From_Rho_Phi(rho, phi, r, dr, beta):
    '''
    the function return the radial velocity dispersion
    even if beta is non zero
    '''

    dphi = libgrid.get_First_Derivative(phi, r)

    if beta==0:
      integrant = rho * dphi
    else:
      integrant = rho * dphi * r**(2*beta)
      
    res = np.array([], np.float32)
    nr = len(dr)

    for i in range(nr):

        ia = i
        ib = nr - 1

        I = get_Integral(integrant, dr, ia=ia, ib=ib)
        res = np.concatenate((res, np.array([I])))

    with np.errstate(divide='ignore', invalid='ignore'): # ignore warning that you divide by zero
      
        if beta==0: 
          sigma = np.where((rho > 0), res / rho, 0)
        else:
          sigma = np.where((rho > 0), res / rho / r**(2*beta), 0)
          sigma[0] = 0  # problem in r=0
    
    sigma = np.sqrt(sigma)

    return sigma


def get_2d_Sigma_From_Rho_Phi(rho, Phi, z, dz):

    res = np.array([], np.float32)
    nr = Phi.shape[0]
    nz = Phi.shape[1]

    for i in range(nr):

        dzPhi = libgrid.get_First_Derivative(Phi[i, :], z)
        integrant = rho[i, :] * dzPhi

        for j in range(nz):

            if (j <= nz / 2):
                ia = 0
                ib = j
                I = -get_Integral(integrant, dz, ia=ia, ib=ib)
                # print "-",ia,ib,I
            else:
                ia = j
                ib = nz - 1
                I = get_Integral(integrant, dz, ia=ia, ib=ib)
                # print "+",ia,ib,I

            res = np.concatenate((res, np.array([I])))

    res.shape = (nr, nz)


    with np.errstate(divide='ignore', invalid='ignore'): # ignore warning that you divide by zero
        sigma_z2 = np.where(rho > 0, res / rho, 0)
    sigma_z = np.where(sigma_z2 > 0, np.sqrt(sigma_z2), 0)

    return sigma_z


def get_2d_Sigma_From_Rho_Phi_Old(m, Phi, z, dz):

    res = np.array([], np.float32)
    nr = Phi.shape[0]
    nz = Phi.shape[1]

    for i in range(nr):

        dzPhi = libgrid.get_First_Derivative(Phi[i, :], z)
        integrant = m[i, :] * dzPhi

        for j in range(nz):

            if (j < (nz - 1) / 2):
                ia = (nz - 1) - j
                ib = nz - 1
            else:
                ia = j
                ib = nz - 1

            #I =  libgrid.get_Integral(integrant,dz,ia=ia,ib=ib)
            I = get_Integral(integrant, dz, ia=ia, ib=ib)

            res = np.concatenate((res, np.array([I])))

    res.shape = (nr, nz)

    with np.errstate(divide='ignore'): # ignore warning that you divide by zero
        sigma_z2 = np.where(m > 0, res / m, 0)
        sigma_z = np.where(sigma_z2 > 0, np.sqrt(sigma_z2), 0)

    return sigma_z



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      libmiyamoto.py
#  brief:     Defines exponential profiles
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
from scipy.special import iv,kv
from scipy.integrate import quad


def Potential(Sigma0,Rd,Rmax,R,z,G=1.0):
    """
    Exponential Potential
    """
    
    def integrand(a,Sigma0,Rd,R,z):
      sqp = np.sqrt(z**2 + (a+R)**2)
      sqm = np.sqrt(z**2 + (a-R)**2)
      I1 = (a+R)/sqp
      I2 = (a-R)/sqm
      I3 = Sigma0*a*kv(1,a/Rd)
      I4 = np.sqrt(R**2-a**2-z**2 + (sqp*sqm))
    
      return (I1-I2)/I4 * I3
     
     
    if type(R)==np.ndarray and type(z)==np.ndarray:
      
      Is = np.zeros(R.shape)
      
      for i in range(R.shape[0]):
        print(i,R.shape[0])
        for j in range(R.shape[1]):

          I = quad(integrand, 0, Rmax, args=(Sigma0,Rd,R[i,j],z[i,j]+1e-5)) 
          
          Is[i,j] = I[0]
                
      return -2*np.sqrt(2)*G*Is    
          
    else:
      I = quad(integrand, 0, Rmax, args=(Sigma0,Rd,R,z)) 
      return -2*np.sqrt(2)*G * I[0]
    
    


def Vcirc(Sigma0,Rd,R,G=1.0):
    """
    Exponential Circular velocity
    """
    
    y = R/(2*Rd)
    B = iv(0,y)*kv(0,y) - iv(1,y)*kv(1,y)
  
    return np.sqrt(4*np.pi*G*Sigma0*Rd * y**2 * B)



def dRVcirc2(Sigma0,Rd,R,G=1.0):
    """
    Radial derivative of the square of the Exponential Circular velocity
    """
    
    y = R/(2*Rd)
    B = iv(0,y)*kv(0,y) - iv(1,y)*kv(1,y)
    
    C1 = kv(0,y) * (iv(1,y) + iv(-1,y))
    C2 = iv(0,y) * (kv(1,y) + kv(-1,y))
    C3 = kv(1,y) * (iv(2,y) + iv(0,y))
    C4 = iv(1,y) * (kv(2,y) + kv(0,y))
    
    C = 0.5*(C1 - C2 - C3 + C4)
  
    return 2*np.pi*G*Sigma0  * ( ( 2*y*B ) + y**2 * C)


def Kappa(Sigma0,Rd,R,G=1.0):
    """
    Radial picycle frequency
    """
    return np.sqrt(Kappa2(Sigma0,Rd,R,G))
    
def Kappa2(Sigma0,Rd,R,G=1.0):
    """
    Square of the Radial picycle frequency
    """
    Vc2 = Vcirc(Sigma0,Rd,R,G)**2
    R2 = R*R
    return dRVcirc2(Sigma0,Rd,R,G)/R + 2*Vc2/R2  


def SurfaceDensity(Sigma0,Rd,R,G=1.0):
    """
    Exponential Density
    """

    return Sigma0*np.exp(-R/Rd)

def Mass(Sigma0,Rd,R,G=1.0):
    """
    Exponential Mass inside R
    """

    return 2*np.pi*Sigma0*Rd**2 * (1- np.exp(-R/Rd)*(1+R/Rd))





 #!/usr/bin/env python3
 ###########################################################################################
 #  package:   pNbody
 #  file:      powerSphericalCutoff.py
 #  brief:     Defines power spherical cut-off profiles
 #  copyright: GPLv3
 #             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
 #             LASTRO - Laboratory of Astrophysics of EPFL
 #  author:    Darwin Roduit <darwin.roduit@epfl.ch>
 #
 # This file is part of pNbody. 
 ###########################################################################################


"""
    Power spherical cut-off model based on the following density:
        \rho(r) = A (\frac{r_1}{r})^\alpha \exp(- \frac{r^2}{r_c^2})
    with A, r_1, \alpha and r_c the parameters of the model.
"""

import numpy as np
from scipy.special import gamma, gammaincc, gammainc

def Potential(alpha, r_c, amplitude, r_1, r, G=1.):
    """
    Potential
    """
    part_1 = - G*CumulativeMass(alpha, r_c, amplitude, r_1, r, G)/r 
    part_2 = - 2*np.pi*G*amplitude *r_1**alpha * r_c**(2-alpha) * gammaincc(1.0-alpha/2, r**2 / r_c**2)*gamma(1.0-alpha/2)
    return part_1+part_2

def CumulativeMass(alpha, r_c, amplitude, r_1, r, G=1.):
    """
    Cumulative mass
    """     
    return 2*np.pi*amplitude*r_1**alpha * r_c**(3.0-alpha) * gammainc(1.5-alpha/2, r**2/r_c**2)*gamma(1.5-alpha/2)    

def Density(alpha, r_c, amplitude, r_1, r, G=1.):
    """
    Density
    """
    return amplitude*(r_1/r)**alpha*np.exp(-(r/r_c)**2)
    

def Vcirc(alpha, r_c, amplitude, r_1, r, G=1.):
    """
    Circular velocity
    """
    M = CumulativeMass(alpha, r_c, amplitude, r_1, r, G=1.)
    return np.sqrt(G*M/r)
     


###########################################################################################
#  package:   Mockimgs
#  file:      spectra.py
#  brief:     filter class
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import sys
import os
import numpy as np

from astropy import units as u
from astropy import constants as c
from scipy import integrate

# absolute magnitude factor 4*np.pi*((10*u.pc.to(u.cm))**2)  = 1.1964951828635063e+40 cm2

class Spectra():
  '''
  Define a spectra class.
  A spectra is characterised by a quantity (luminosity or transmission) given as a function
  of the wavelength (by default in angstrom)
  '''

  def __init__(self):  
    pass
    
  
  def resample(self,wavelength=None,lmin=1e1,lmax=1e7,n=1000000,units=u.angstrom):
    """
    resample data over a different wavelength range
    """
    
    if wavelength is None:
      logmin = np.log10(lmin)
      logmax = np.log10(lmax)
      wavelength = np.logspace(logmin,logmax,n)
    
    # interpolate data
    if type(self.data) is np.ndarray:
      data = np.interp(wavelength,self.wavelength.value,self.data)
      self.data       = data
    else:
      data = np.interp(wavelength,self.wavelength.value,self.data.value)
      self.data       = data*  self.data.unit
    
    self.wavelength = wavelength*units
    
    
    
    
  def get_wavelength(self,units=u.angstrom):
    """
    get the wavelength
    """
    return self.wavelength.to(units)
  
  
  def get_data(self,units=u.erg/u.s/u.angstrom):
    """
    get the data
    """
    if units is None:
      return self.data
    else:  
      return self.data.to(units)    
  
  def get_luminosity(self):
    """
    compute the luminosity, i.e., the integral under the data curve
    """
    from scipy import integrate
    I = integrate.simpson(self.data, self.wavelength)  
    return I
  

  
class SED(Spectra):
  '''
  Define a Spectral Energy Distribution class.
  The class is inherited from the Spectra class
  '''  
  
  def __init__(self,sed,wavelength):  
    
    if type(sed)   is np.ndarray:
      self.data        = sed * u.erg/u.s/u.angstrom
    elif type(sed) is u.quantity.Quantity:
      self.data        = sed
    else:
      raise(ValueError)
    
    if type(wavelength)   is np.ndarray:
      self.wavelength        = wavelength * u.angstrom
    elif type(wavelength) is u.quantity.Quantity:
      self.wavelength        = wavelength
    else:
      raise(ValueError)
  
  def mean_ST_flux(self,f):
    """
    compute the mean flux using the Willmer 2018 prescription
    f is a 
    """

    if not np.all(self.wavelength==f.wavelength):
      raise ValueError('self.wavelength must be identical to f.wavelength')
     
    F = self.absolute_flux
    R = f.data
    l = self.wavelength
  
    I1 = integrate.simpson(F*R*l,l)
    I2 = integrate.simpson(  R*l,l)
  
    flm = I1/I2 * u.erg/u.s/u.cm**2/u.AA
    
    return flm
    
  def mean_AB_flux(self,f):
    """
    compute the mean flux using the Willmer 2018 prescription
    f is the response 
    """

    if not np.all(self.wavelength==f.wavelength):
      raise ValueError('self.wavelength must be identical to f.wavelength')
     
    F = self.absolute_flux
    R = f.data
    l = self.wavelength
    
    F  = F.to(u.erg/(u.s*u.cm**2*u.cm)).value     # to erg/(s*cm**2*cm)
    l  = l.to(u.cm).value                         # to cm
    cc = c.c.to(u.cm/u.s).value                   # to cm/s

    I1 = integrate.simpson(F*R*l,l)
    I2 = integrate.simpson(  R*cc/l,l)  
    flm = I1/I2  * u.erg/u.s/u.cm**2/u.Hz        # give new appropriate units    
    
    return flm    
    
    
    
  
  def STMagnitude(self,f):
    """
    compute ST magnitude
    f is the filter "SED"
    See Bessell and Murphy 2012 for the origin of 21.10
    """

    # compute the mean flux
    flm = self.mean_ST_flux(f)

    # prevent 0 flux
    flm = np.where(flm==0,1e-50*flm.unit,flm)
    
    # compute the AB magnitude
    mag = -2.5* np.log10(flm.value) - 21.10
      
    return mag
    
  def ABMagnitude(self,f):
    """
    convert the flux in [u.erg/u.s/u.cm**2/u.AA]
    in AB magnitude using the filter "f"
    """

    # compute the mean flux (integral of f with the filter response)
    flm = self.mean_AB_flux(f)
    
    # prevent 0 flux
    flm = np.where(flm==0,1e-50*flm.unit,flm)

    # compute the AB magnitude
    mag = -2.5* np.log10(flm.value) - 48.6
      
    return mag

  def BolometricLuminosity(self):
    """
    from the intrinsic luminosity [erg/s/angstrom]
    compute the bolometric Luminosity [erg/s]
    """

    l = self.get_wavelength()     # in Angstrom
    f = self.data                 # in erg /(s Angstrom)
    
    funit = (1*u.erg/u.s/u.Angstrom).cgs.unit
    lunit = (1*u.Angstrom).cgs.unit
    
    if f.cgs.unit != funit:
      raise ValueError('f must be in units of %s'%funit)
    if l.cgs.unit != lunit:
      raise ValueError('l must be in units of %s'%lunit)
      
    I = integrate.simpson(f,l)    # erg/s
    I = I*u.erg/u.s    
    
    return I
    
    
  def Luminosity(self,f):
    """
    from the intrinsic luminosity [erg/s/angstrom]
    compute the luminosity [erg/s] in a given filter f
    """
    SED = self.multiply_with_filter(f)
    I = SED.BolometricLuminosity()
    return I


  def toAbsoluteFlux(self):
    """
    copy the current intensity stored in the .data variable to the
    .absolute_flux variable
    u.erg/u.s/u.cm**2/u.AA
    """
    self.absolute_flux = self.data
            
  
  
  def computeAbsoluteFlux(self):
    """
    convert intrinsic luminosity [erg/s/angstrom] 
    into absolute flux [u.erg/u.s/u.cm**2/u.AA],
    i.e. move the source at 10pc.
    """
    
    # compute the flux at a distance of 10 pc
    d = 10*u.pc
    d = d.to(u.cm)

    self.absolute_flux = self.data/(4*np.pi * d**2)
    self.absolute_flux.to(u.erg/u.s/u.cm**2/u.AA)
    
    
  
  
  def multiply_with_filter(self,f):
    """
    multiply the SED with a given filter
    both must have the same wavelength range
    """
    
    if not np.all(self.wavelength==f.wavelength):
      raise ValueError('self.wavelength must be identical to f.wavelength')
    
    data = self.data * f.data
    
    return SED(data,self.wavelength)
    
    

  def get_wavelength(self,units=u.angstrom):
    """
    get the wavelength
    """
    return self.wavelength.to(units)
  
  def get_sed(self,units=u.erg/u.s/u.angstrom):
    """
    get the sed
    """
    return self.get_data(units=units)
    
    
    
    



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      pm.py
#  brief:     Defines point mass profile
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


"""
point mass model
"""

import numpy as np


def Potential(M, r, G=1.):
    """
    Potential
    """
    return -G * M / np.sqrt(r**2)


def dPotential(M, r, G=1.):
    """
    first derivative of Potential
    """
    return G * M * r * (r**2)**(-3. / 2.)

def Density(M, r, G=1.):
    """
    Density
    """
    if type(r) == np.ndarray:
      return np.zeros(len(r)) 
    else:
      return 0
    
    
def Vcirc(M, r, G=1.):
    """
    circular velocity
    """
    return np.sqrt(G * M / r)





#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      geometry.py
#  brief:     Geometry transformation/computation
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np


def norm(x):
    """
    return the norm of vector x
    """
    return np.sqrt(x[0]**2 + x[1]**2 + x[2]**2)


def rotate(x, angle=0, axis=[1, 0, 0], point=[0, 0, 0]):
    """
    Rotate the positions and/or the velocities of the object around a specific axis
    with respect to a specific point

    angle : rotation angle in radian
    axis  : [x,y,z] : around this axis
    point : [x,y,z] : rotation origin

    use the euler rotation matrix
    """

    # center point
    x = x - point

    # construction of the rotation matrix
    norm = np.sqrt(axis[0]**2 + axis[1]**2 + axis[2]**2)
    if norm == 0:
        return x
    sn = np.sin(-angle / 2.)

    e0 = np.cos(-angle / 2.)
    e1 = axis[0] * sn / norm
    e2 = axis[1] * sn / norm
    e3 = axis[2] * sn / norm

    a = np.zeros((3, 3), float)
    a[0, 0] = e0**2 + e1**2 - e2**2 - e3**2
    a[1, 0] = 2. * (e1 * e2 + e0 * e3)
    a[2, 0] = 2. * (e1 * e3 - e0 * e2)
    a[0, 1] = 2. * (e1 * e2 - e0 * e3)
    a[1, 1] = e0**2 - e1**2 + e2**2 - e3**2
    a[2, 1] = 2. * (e2 * e3 + e0 * e1)
    a[0, 2] = 2. * (e1 * e3 + e0 * e2)
    a[1, 2] = 2. * (e2 * e3 - e0 * e1)
    a[2, 2] = e0**2 - e1**2 - e2**2 + e3**2
    a = a.astype(float)

    # multiply x
    x = np.dot(x, a)

    # decenter point
    return x + point


def align(x, axis1=[1, 0, 0], axis2=[0, 0, 1], point=[0, 0, 0]):
    """
    Rotate the object around point in order to align the axis 'axis1' with the axis 'axis2'.

    axis1 : [x,y,z]
    axis2 : [x,y,z]
    """

    a1 = np.array(axis1, float)
    a2 = np.array(axis2, float)

    a3 = np.array([0, 0, 0], float)
    a3[0] = a1[1] * a2[2] - a1[2] * a2[1]
    a3[1] = a1[2] * a2[0] - a1[0] * a2[2]
    a3[2] = a1[0] * a2[1] - a1[1] * a2[0]

    n1 = np.sqrt(a1[0]**2 + a1[1]**2 + a1[2]**2)
    n2 = np.sqrt(a2[0]**2 + a2[1]**2 + a2[2]**2)
    angle = np.arccos(np.inner(a1, a2) / (n1 * n2))

    return rotate(x, angle=angle, axis=a3, point=point)


def expose(obj, obs, eye=None, dist_eye=None, foc=None):
    """
    Rotate and translate the object in order to be seen as if the
    observer was in x0, looking at a point in xp.

    obj	      : object array to expose
    obs         : representation of the observer
    eye	      : 'right' or 'left'
    dist_eye    : distance between eyes (separation = angle))
    foc         : focal

    """

    # first : put x0 at the origin
    obj = obj - obs[0]
    obs = obs - obs[0]

    # second : anti-align e1 with z
    obj = align(obj, axis1=obs[1], axis2=[0, 0, -1])
    obs = align(obs, axis1=obs[1], axis2=[0, 0, -1])

    # third : align e3 with y
    obj = align(obj, axis1=obs[3], axis2=[0, 1, 0])

    # fourth if eye is defined
    if eye == 'right':
        if foc is not None:
            Robs = foc
        else:
            Robs = np.sqrt((obs[0][0] - obs[1][0])**2 + (obs[0]
                                                      [1] - obs[1][1])**2 + (obs[0][2] - obs[1][2])**2)
        phi = -np.arctan(dist_eye)
        obj = rotate(obj, -phi, axis=[0, 1, 0], point=[0, 0, -Robs])
    elif eye == 'left':
        if foc is not None:
            Robs = foc
        else:
            Robs = np.sqrt((obs[0][0] - obs[1][0])**2 + (obs[0]
                                                      [1] - obs[1][1])**2 + (obs[0][2] - obs[1][2])**2)
        phi = -np.arctan(dist_eye)
        obj = rotate(obj, +phi, axis=[0, 1, 0], point=[0, 0, -Robs])

    return obj


#  def perspective(r_obs=100., foc=50., view='xz'):
#      """
#      Project the N-body model in order to get a perspective view.
#
#      r_obs = distance of the observer to the looking point
#      foc	= focal
#      view  = 'xz' , 'xy' , 'yz'
#      """
#      from pNbody.nbodymodule import perspective as perspec
#
#      if view == 'xz':
#          view = 1
#      elif view == 'xy':
#          view = 2
#      elif view == 'yz':
#          view = 3
#      elif view != 'xz'and view != 'xy'and view != 'yz':
#          view = 1
#
#      return perspec(pos, r_obs, foc, view)


def frustum(pos, clip, size):
    """
    Project using a frustrum matrix

    clip = near and far planes
    size = size of the box
    """

    tmp = np.ones((pos.shape[0], pos.shape[1] + 1), float)
    tmp[:, :-1] = pos
    pos = tmp

    n = float(clip[0])
    f = float(clip[1])
    l = -float(size[0])
    r = float(size[0])
    b = -float(size[1])
    t = float(size[1])

    # frustum
    m1 = np.zeros((4, 4), float)

    m1[0, 0] = 2. * n / (r - l)
    m1[1, 0] = 0.
    m1[2, 0] = (r + l) / (r - l)
    m1[3, 0] = 0.

    m1[0, 1] = 0.
    m1[1, 1] = 2. * n / (t - b)
    m1[2, 1] = (t + b) / (t - b)
    m1[3, 1] = 0.

    m1[0, 2] = 0.
    m1[1, 2] = 0.
    m1[2, 2] = -(f + n) / (f - n)
    m1[3, 2] = -(2. * f * n) / (f - n)

    m1[0, 3] = 0.
    m1[1, 3] = 0.
    m1[2, 3] = -1
    m1[3, 3] = 0.

    m1 = m1.astype(float)

    posc = np.dot(pos, m1)
    posc = np.transpose(np.transpose(posc) / posc[:, 3])[:, :-1]

    return posc


def ortho(pos, clip, size):
    """
    Project using an ortho matrix

    clip = near and far planes
    size = size of the box
    """

    tmp = np.ones((pos.shape[0], pos.shape[1] + 1), float)
    tmp[:, :-1] = pos
    pos = tmp

    n = float(clip[0])
    f = float(clip[1])
    l = -float(size[0])
    r = float(size[0])
    b = -float(size[1])
    t = float(size[1])

    # ortho
    m1 = np.zeros((4, 4), float)

    m1[0, 0] = 2. / (r - l)
    m1[1, 0] = 0.
    m1[2, 0] = 0.
    m1[3, 0] = -(r + l) / (r - l)

    m1[0, 1] = 0.
    m1[1, 1] = 2. / (t - b)
    m1[2, 1] = 0
    m1[3, 1] = -(t + b) / (t - b)

    m1[0, 2] = 0.
    m1[1, 2] = 0.
    m1[2, 2] = -2. / (f - n)
    m1[3, 2] = -(f + n) / (f - n)

    m1[0, 3] = 0.
    m1[1, 3] = 0.
    m1[2, 3] = 0.
    m1[3, 3] = 1.

    m1 = m1.astype(float)

    posc = np.dot(pos, m1)
    posc = np.transpose(np.transpose(posc) / posc[:, 3])[:, :-1]

    return posc


def viewport(xn, shape=None):
    """
    viewport transformation

    xn    = position (output from frustum or ortho)
    shape = shape of the image
    """

    x = xn[:, 0]
    y = xn[:, 1]
    z = xn[:, 2]

    if shape is None:
        wx = 1
        wy = 1
    else:
        wx = shape[0]
        wy = shape[1]

    winx = (x + 1) / 2. * wx
    winy = -(y - 1) / 2. * wy
    winz = (z + 1) / 2.

    return np.transpose(np.array([winx, winy, winz]))


def inv_viewport(xw, yw, zw, shape):
    """
    viewport transformation

    xn    = position (output from frustum or ortho)
    shape = shape of the image
    """

    wx = shape[0]
    wy = shape[1]

    x = 2. * xw / wx - 1
    y = -2. * yw / wy + 1
    z = 2. * zw - 1

    return x, y, z


def boxcut(pos, args):
    """
    Return only particles that are inside box 1:1:1
    """

    c = np.logical_not((np.fabs(pos[:, 0]) > 1) | (
        np.fabs(pos[:, 1]) > 1) | (np.fabs(pos[:, 2]) > 1))
    pos = np.compress(c, pos, axis=0)
    newargs = ()
    for arg in args:
        newargs = newargs + (np.compress(c, arg, axis=0),)

    return pos, newargs


def boxcut_segments(pos, args):
    """
    Return only particles that are inside box 1:1:1
    """

    n = len(pos)
    p = int(n / 2)

    newpos = np.zeros(pos.shape)
    cs = np.zeros(len(pos))
    NP = 0

    for j in range(p):
        i = 2 * j

        f0 = False
        f1 = False
        if (np.fabs(pos[i, 0]) > 1) or (
                np.fabs(pos[i, 1]) > 1) or (np.fabs(pos[i, 2]) > 1):
            f0 = True
        if (np.fabs(pos[i + 1, 0]) > 1) or (np.fabs(pos[i + 1, 1])
                                         > 1) or (np.fabs(pos[i + 1, 2]) > 1):
            f1 = True

        if f0 and f1:
            # both outside : remove the segment
            # print "remove"
            pass
        elif f0 and not f1:
            # 0 outside 1 inside : change 0
            # print "change 0"
            newpos[i] = pos[i]
            newpos[i + 1] = pos[i + 1]
            cs[i] = 0		# remove
            cs[i + 1] = 0
            NP += 2
        elif not f0 and f1:
            # 0 inside 1 outside : change 1
            # print "change 1"
            newpos[i] = pos[i]
            newpos[i + 1] = pos[i + 1]
            cs[i] = 0		# remove
            cs[i + 1] = 0
            NP += 2
        else:
            # both inside  : keep
            newpos[i] = pos[i]
            newpos[i + 1] = pos[i + 1]
            cs[i] = 1
            cs[i + 1] = 1
            NP += 2

    pos = np.compress(cs, newpos, axis=0)

    newargs = ()
    for arg in args:
        newargs = newargs + (np.compress(cs, arg, axis=0),)

    return pos, newargs


def get_obs(x0=[0., -50., 0.], xp=[0., 0., 0.], alpha=0, view='xz', r_obs=50):
    """
    From some parameters, return an obs matrix

    {x0,xp,alpha} or {xz}

    """

    # use view and r_obs
    if (x0 is None or xp is None or alpha is None) and (
            view is not None and r_obs is not None):
        if view == 'xz':
            e0 = np.array([0, -r_obs, 0], float)
            e1 = np.array([0, 0, 0], float)
            e2 = np.array([-1, -r_obs, 0], float)
            e3 = np.array([0, -r_obs, 1], float)
        elif view == 'xy':
            e0 = np.array([0, 0, r_obs], float)
            e1 = np.array([0, 0, 0], float)
            e2 = np.array([-1, 0, r_obs], float)
            e3 = np.array([0, 1, r_obs], float)
        elif view == 'yz':
            e0 = np.array([r_obs, 0, 0], float)
            e1 = np.array([0, 0, 0], float)
            e2 = np.array([r_obs, -1, 0], float)
            e3 = np.array([r_obs, 0, 1], float)
        else:
            e0 = np.array([0, -r_obs, 0], float)
            e1 = np.array([0, 0, 0], float)
            e2 = np.array([-1, -r_obs, 0], float)
            e3 = np.array([0, -r_obs, 1], float)

        # create the matrix
        obs = np.array([e0, e1, e2, e3], float)

        # center xp
        if xp is not None:
            obs = obs + xp

    # use x0,xp,alpha
    else:

        x0 = np.array(x0, float)
        xp = np.array(xp, float)

        Robs = np.sqrt((x0[0] - xp[0])**2 + (x0[1] - xp[1])
                    ** 2 + (x0[2] - xp[2])**2)

        e0 = np.array([0, 0, 0], float)
        e1 = np.array([Robs, 0, 0], float)
        e2 = np.array([0, 1, 0], float)
        e3 = np.array([0, 0, 1], float)

        # create the matrix
        obs = np.array([e0, e1, e2, e3], float)

        #####################

        yp = xp - x0
        rxy = np.sqrt(yp[0]**2 + yp[1]**2)
        z = yp[2]

        a = np.arctan2(yp[1], yp[0])

        if rxy != 0:
            b = np.arctan(z / rxy)
        else:
            b = 0.

        # rotate azimuth
        obs = rotate(obs, angle=a, axis=[0, 0, 1])

        # rotate elevation
        obs = rotate(obs, angle=-b, axis=obs[2])

        # rotate (xa around e1)
        obs = rotate(obs, angle=alpha, axis=obs[1])

        # translate to x0
        obs = obs + x0


    # # old version (bad)
    #
    # # create the matrix
    # obs =  np.array([e0,e1,e2,e3],float)
    #
    # # rotate (align e1 with xp-x0 )
    # obs = align(obs,axis1=obs[1],axis2=xp-x0)
    #
    # # rotate (xa around e1)
    # obs = rotate(obs,angle=alpha,axis=obs[1])
    #
    # # translate to x0
    # obs = obs + x0



        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # !!! e2 may not be parallel to the plane z=0 !!!
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    return obs



#!/usr/bin/env python3

###########################################################################################
#  package:   pNbody
#  file:      libbruzual.py
#  brief:     bruzual luminosities
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
###########################################################################################

import os
import numpy as np

from .libSSPluminosity import SSPLuminosities


#####################################################
def read_ascii(
        file,
        columns=None,
        lines=None,
        dtype=float,
        skipheader=False,
        cchar='#'):
    #####################################################
    """[X,Y,Z] = READ('FILE',[1,4,13],lines=[10,1000])
    Read columns 1,4 and 13 from 'FILE'  from line 10 to 1000
    into array X,Y and Z

    file is either fd or name file

    """

    def RemoveComments(l):
        if l[0] == cchar:
            return None
        else:
            return l

    def toNumList(l):
        return list(map(dtype, l))

    if isinstance(file, str):
        f = open(file, 'r')
    else:
        f = file

    # read header while there is one
    while True:
        fpos = f.tell()
        header = f.readline()
        if header[0] != cchar:
            f.seek(fpos)
            header = None
            break
        else:
            if skipheader:
                header = None
            else:
                # create dict from header
                header = str.strip(header[2:])
                elts = str.split(header)
                break

    """
  # read header if there is one
  header = f.readline()
  if header[0] != cchar:
    f.seek(0)
    header = None
  else:
    if skipheader:
      header = None
    else:
      # create dict from header
      header = string.strip(header[2:])
      elts = string.split(header)
  """

    # now, read the file content
    lines = f.readlines()

    # remove trailing
    lines = list(map(str.strip, lines))

    # remove comments
    #lines = map(RemoveComments, lines)

    # split
    lines = list(map(str.split, lines))

    # convert into float
    lines = list(map(toNumList, lines))

    # convert into array
    lines = np.array(list(map(np.array, lines)))

    # np.transpose
    lines = np.transpose(lines)

    if header is not None:
        iobs = {}
        i = 0
        for elt in elts:
            iobs[elt] = i
            i = i + 1

        vals = {}
        for key in list(iobs.keys()):
            vals[key] = lines[iobs[key]]

        return vals

    # return
    if columns is None:
        return lines
    else:
        return lines.take(axis=0, indices=columns)


class BruzualLuminosities(SSPLuminosities):

    def __init__(self, directory):

        Z0 = 0.02

        self.Ids = [22, 32, 42, 52, 62, 72]
        self.Zs = np.array([0.0001, 0.0004, 0.004, 0.008, 0.02, 0.05])
        self.Zs = np.log10(self.Zs / Z0)

        self.directory = directory

        # read files and crate self.data
        self.Read()

        # create the matrix
        self.CreateMatrix()

    def Read(self):
        """
        read files and create a data table
        """

        self.data = {}

        for Id, Z in zip(self.Ids, self.Zs):
            name = os.path.join(
                self.directory,
                "bc2003_hr_m%d_salp_ssp.1color" %
                (Id))
            data = read_ascii(name, skipheader=True, cchar='#')
            self.data[Z] = data

    def CreateMatrix(self):
        """
        from data extract
        metalicites (zs)
        ages (ages)
        and ML (vs)
        """

        # create borders
        Zs = self.Zs
        Ages = (10**self.data[Zs[0]][0]) / 1e9

        MatLv = np.zeros((len(Zs), len(Ages)))

        for iZ, z in enumerate(Zs):
            mv = self.data[z][4]
            mv0 = 4.83  # ok
            Lv = 10**(-(mv - mv0) / 2.5)
            MatLv[iZ, :] = Lv

        self.MatLv = MatLv
        self.Ages = Ages
        self.Zs = Zs



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      BastI.py
#  brief:     
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
import os
from scipy import interpolate


database_directory=None
default_keys=None



###########################################################
# 
# Isochrone block class
#
###########################################################


class IsochroneBlock():
  """
  This class contains isochrone data, i.e., data for a list of stars of
  different masses, but with the same metallicity and age.
  """
    
  def __init__(self,data=None,keys=None,MH=None,Age=None,Mmin=None,Mmax=None):
    
    self.data = data
    self.keys = keys
    
    self.Mmin = Mmin    # minimum stellar mass
    self.Mmax = Mmax    # maximal stellar mass
    self.MH   = MH      # metallicity
    self.Age  = Age     # age

  
  def get(self,key):
    "get the values corresponding to a key (a column)"
    return self.data[:,self.keys[key]]




###########################################################
# 
# The main Isochrone class
#
###########################################################


class IsochronesGrid():
  """
  
  self.M is a numpy 2D matrix containing isochrones data (self.M[i,j]) for 
  a given metallicity and age.
  
  The Metallicity and Ages considered for the matrix is:
  
  self.binsMH  : in log10[M/H]
  self.binsAge : in Myr
  
  """
  
  def __init__(self,dirname=database_directory,default_keys=default_keys,filename=None):
    
    # make some parmeters global
    self.dirname      = dirname
    self.filename     = filename
    self.default_keys = default_keys
    
    # init the data base (the grid)
    self.Init() 
     



  def organizeDB(self):
    """
    Create the self.M matrix from the isochrones blocks
    """
    
    # compute min/max MH and Ages
    self.binsMH = []
    self.binsAge = []
    
    for mb in self.MBs:
      if mb.MH not in self.binsMH:
        self.binsMH.append(mb.MH)
      if mb.Age not in self.binsAge:
        self.binsAge.append(mb.Age)

    self.binsAge = np.array(self.binsAge)
    self.binsMH  = np.array(self.binsMH)
    self.binsAge.sort()
    self.binsMH.sort()
    
    self.M = np.empty((len(self.binsAge),len(self.binsMH)),dtype="O")
    
    # loop again and fill the matrix
    for mb in self.MBs:
      i = self.binsAge.tolist().index(mb.Age)
      j = self.binsMH.tolist().index(mb.MH)
      self.M[i,j] = mb

  def FillDBGaps(self):
    """
    Fill missing isochrones with the nearest one
    """
    pass
    
  
  
    
  def getMHbinsNumber(self):
    return len(self.binsMH)
    
  def getAgebinsNumber(self):
    return len(self.binsAge)     
  
  
  def getAgeIndex(self,Age):  
    '''
    determine the index for a given Age
    '''
    return np.fabs(self.binsAge - Age).argmin()

  def getAgeIndexes(self,Age):  
    '''
    determine the indexes bracketing a given Age
    '''
    idx = np.fabs(self.binsAge - Age).argmin()
    dd = Age - self.binsAge[idx]
    n = len(self.binsAge)-1
   
    if dd < 0:
      idx1 = idx - 1
      idx2 = idx
    else:
      idx2 = idx + 1
      idx1 = idx  

    idx1 = np.clip(idx1,0,n)
    idx2 = np.clip(idx2,0,n)
   
    return idx1,idx2

  def getMHIndex(self,MH):  
    '''
    determine the index for a given MH
    '''
    return np.fabs(self.binsMH - MH).argmin()

  def getMHIndexes(self,MH):  
    '''
    determine the indexes bracketing a given MH
    '''
    idx = np.fabs(self.binsMH - MH).argmin()
    dd = MH - self.binsMH[idx]
    n = len(self.binsMH)-1
   
    if dd < 0:
      idx1 = idx - 1
      idx2 = idx
    else:
      idx2 = idx + 1
      idx1 = idx  

    idx1 = np.clip(idx1,0,n)
    idx2 = np.clip(idx2,0,n)
   
    return idx1,idx2


  def getMaxStellarMass(self,Age,MH): 
    '''
    for a given Age and MH, get the maximal stellar mass
    '''
    # get the indexes of the table
    i = self.getAgeIndex(Age)
    j = self.getMHIndex(MH)
    # get the table
    M = self.M[i,j]  
    return M.Mmax  
    
  
  def getMassIndexes(self,Mass,i,j):
    '''
    From i (Age) and j (MH), and Mass the mass, 
    return the two indexes bracketing the mass. 
    '''
    # get the table
    M = self.M[i,j]     

    Masses = M.data[:,M.keys['initial_mass']]
    n = len(Masses)-1
    
    # a numpy array
    if type(Mass)==np.ndarray:
      idx2 = Masses.searchsorted(Mass)
      idx1 = idx2-1
      return idx1, idx2
       
    # a scalar  
    else:
      # here we need to find two masses
      idx = np.fabs(Masses - Mass).argmin()
      dd = Mass - Masses[idx]

      if dd < 0:
        idx1 = idx - 1
        idx2 = idx
      else:
        idx2 = idx + 1
        idx1 = idx  
      
      idx1 = np.clip(idx1,0,n)
      idx2 = np.clip(idx2,0,n)
      
      return idx1,idx2
      
      

  def interpolateMassAndValue(self,Mass,i,j,key):
    '''
    From i (Age) and j (MH), return the interpolated key
    for a given mass Mass
    '''
    
    # get the tables and mass indexes
    M = self.M[i,j] 
    
    k1,k2 = self.getMassIndexes(Mass,i,j)
   
    # get values
    M1 = M.data[k1,M.keys['initial_mass']]
    M2 = M.data[k2,M.keys['initial_mass']]
    V1 = M.data[k1,M.keys[key]]
    V2 = M.data[k2,M.keys[key]]
   
    # linear interpolation
    if type(Mass)==np.ndarray:
      V = np.where(V1!=V2,(M1 - Mass)/(M1-M2) * (V2-V1) + V1,V1)      
    else:  
      if M1 != M2:
        V = (M1 - Mass)/(M1-M2) * (V2-V1) + V1 
      else:
        V = V1
      
    return V    



  def get(self,Age,MH,Mass,key=None,mode="masslininterp"):
    """
    For a given Age, MH, Mass, return the value given
    by the key key.
    
    mode provides different ways of interpolating tables.
    masslininterp interpolates over masses and  
    """
    
    if mode=="nearest":
      '''
      use nearest value for the Age, MH and Mass
      '''
    
      # get the indexes of the table
      i = self.getAgeIndex(Age)
      j = self.getMHIndex(MH)
      # get the table
      M = self.M[i,j]
    
    
      Masses = M.data[:,M.keys['initial_mass']]
      idx = np.fabs(Masses - Mass).argmin()
    
      if key is not None:
        val = M.data[idx,M.keys[key]]   
        
      return val
      
    if mode=="masslininterp":
      '''
      use nearest value for the Age and MH,
      interpolate the mass
      '''
      # get the indexes of the table
      i = self.getAgeIndex(Age)
      j = self.getMHIndex(MH)
      # get the table
      #M = self.M[i,j]
      
      # interpolate the mass
      v = self.interpolateMassAndValue(Mass,i,j,key)
      return v    
    
    
    
    
    elif mode=="Agelininterp":
      
     # get the indexes and values
     i1,i2 = self.getAgeIndexes(Age)
     Age1  = self.binsAge[i1]
     Age2  = self.binsAge[i2] 
      
     # get the nearest metallicity
     j = self.getMHIndex(MH)
     
     # interpolate the Age
     v1 = self.interpolateMassAndValue(Mass,i1,j,key)
     v2 = self.interpolateMassAndValue(Mass,i2,j,key)
     val = (Age1 - Age)/(Age1-Age2) * (v2-v1) + v1 
     return val        


    elif mode=="MHlininterp":
      
    # interpolate only on MH
      
     # get the indexes and values
     j1,j2 = self.getMHIndexes(MH)
     MH1  = self.binsMH[j1]
     MH2  = self.binsMH[j2] 
     
     # get the nearest metallicity
     i = self.getAgeIndex(Age)

     
     # special case : if we are between -2.5 and -2 in MH, 
     # the isochrones are strongly different for the high mass end (wd)
     # do a nearest interpolation
     if j1==3 and j2==4:
       # get the indexes of the table
       i = self.getAgeIndex(Age)
       j = self.getMHIndex(MH)
       # get the table
       M = self.M[i,j]
       Masses = M.data[:,M.keys['initial_mass']]
       idx = np.fabs(Masses - Mass).argmin()
       if key is not None:
         val = M.data[idx,M.keys[key]]   
       return val
     
     else:
       # interpolate the MH
       v1 = self.interpolateMassAndValue(Mass,i,j1,key)
       v2 = self.interpolateMassAndValue(Mass,i,j2,key)
       val = (MH1 - MH)/(MH1-MH2) * (v2-v1) + v1 
       return val  

    
    
    elif mode=="lininterp":
      
      # get the indexes and values
      i1,i2 = self.getAgeIndexes(Age)
      Age1  = self.binsAge[i1]
      Age2  = self.binsAge[i2]
            
      j1,j2 = self.getMHIndexes(MH)
      MH1   = self.binsMH[j1]
      MH2   = self.binsMH[j2]   
      
      # MH1=MH2 and Age1=Age2 -> not interpolation
      if j1==j2 and i1==i2: 
        v = self.interpolateMassAndValue(Mass,i1,j1,key)
        return v
        
      # MH1=MH2 -> interpolate only in Age
      if j1==j2: 
        v1 = self.interpolateMassAndValue(Mass,i1,j1,key)
        v2 = self.interpolateMassAndValue(Mass,i2,j1,key)
        val = (Age1 - Age)/(Age1-Age2) * (v2-v1) + v1 
        return val
      
      # Age1=Age2 -> interpolate only in MH  
      if i1==i2:
        v1 = self.interpolateMassAndValue(Mass,i1,j1,key)
        v2 = self.interpolateMassAndValue(Mass,i1,j2,key)
        val = (MH1 - MH)/(MH1-MH2) * (v2-v1) + v1 
        return val
      
      
      # special case : if we are between -2.5 and -2 in MH, 
      # the isochrones are strongly different for the high mass end (wd)
      # we then only interpolate in Age
      if j1==3 and j2==4:
        # get the nearest metallicity
        j = self.getMHIndex(MH)
        v1 = self.interpolateMassAndValue(Mass,i1,j,key)
        v2 = self.interpolateMassAndValue(Mass,i2,j,key)
        val = (Age1 - Age)/(Age1-Age2) * (v2-v1) + v1 
        return val        
        
        
        
      # here we can perform a bilinear interpolation
      
      # get the four values
      v11 = self.interpolateMassAndValue(Mass,i1,j1,key)
      v12 = self.interpolateMassAndValue(Mass,i1,j2,key)
      v21 = self.interpolateMassAndValue(Mass,i2,j1,key)
      v22 = self.interpolateMassAndValue(Mass,i2,j2,key)
      
      #if np.isfinite(v11) == False:
        #print(v11,Age1,MH1,Mass)
        #exit()

      # if np.isfinite(v12) == False:
        # print(v12,Age1,MH2,Mass)
        # exit()
        
      # if np.isfinite(v21) == False:
        # print(v21,Age2,MH1,Mass)
        # exit()
        
      # if np.isfinite(v22) == False:
        # print(v22,Age2,MH2,Mass)
        # exit()

      # bilinear interpolation
      x = np.array([Age1,Age2])
      y = np.array([MH1,MH2])
      xx, yy = np.meshgrid(x, y)
      z = np.array([[v11,v21],[v12,v22]])
      f = interpolate.interp2d(x, y, z)
      val =  f(Age,MH)[0]
      
      # in case something goes wrong
      if np.isfinite(val) == False:
        print("!!! bad value : val = %g"%val)
        print(Age,MH,Mass)
        print(Age1,Age2,MH1,MH2)
        exit()
            
            
      return val


  def getKeys(self):
    """
    return the list of the keys
    """
    return self.keys

  def checkKeyExists(self,key):
    """
    check if the key key is the key list
    """
    for k in self.getKeys():
      if k == key:
        return True
    
    raise NameError("key %s does not exists in DB.\nUse one given in the following list: \n%s"%(key,str(list(self.getKeys().keys()))))  






#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      CMD.py
#  brief:     
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
import tarfile
import glob
import os
from pNbody.Isochrones.main import  IsochronesGrid, IsochroneBlock



def read_CMD(filename,MBs=[],default_keys=None):
  """
  read one CMD file
  http://stev.oapd.inaf.it/cmd
  
  and return it content as an array
  
  filename      : name of the file to read
  default_keys  : default keys to read. If None, ignore

  """
  
  # find the header and parse it
  f = open(filename,'r')
  while(True):
    line = f.readline()
    if line[:6] == "# Zini":
      break
  f.close()
  
  keys = line[1:].split()
  # change Mini to initial_mass
  idx = keys.index("Mini")
  keys[idx] = "initial_mass"
  
  # make a dict
  keys = {x: i for i,x in enumerate(keys)}
  

  if default_keys is not None:
    new_keys = {}
    tmp_keys = []
    for i,k in enumerate(default_keys):
      new_keys[k] = i
      tmp_keys.append(keys[k])
    keys = new_keys  
  else:
    tmp_keys = []
    for i,k in enumerate(keys):
      tmp_keys.append(keys[k])

    
  # read the complete data
  data = np.loadtxt(filename,comments="#",usecols=tmp_keys)
  
  MH     = data[:,keys["MH"]]
  Age    = 10**data[:,keys["logAge"]]/1e6  # in Myr
  
  # find the bins
  binsMH  = np.unique(MH)
  binsAge = np.unique(Age)
  
  
  M = np.empty((len(binsAge),len(binsMH)),dtype="O")
  
  
  for i in range(len(binsAge)):
    for j in range(len(binsMH)):

      # create a mist block
      dataij = data[(MH==binsMH[j]) & (Age==binsAge[i])]
      
      MB = IsochroneBlock(data=dataij,keys=keys)
      
      # min/max stellar mass in solar masses
      MB.Mmin = min(MB.get("initial_mass"))
      MB.Mmax = max(MB.get("initial_mass"))
      MB.MH   = binsMH[j]    
      # age in Myr
      MB.Age  = binsAge[i]/1e6       
      
      M[i,j] = MB
              
  return M,binsMH,binsAge,keys
  


###########################################################
# 
# Isochrone Grid
#
###########################################################

class Grid(IsochronesGrid):
  
  def Init(self):

    self.M,self.binsMH,self.binsAge,self.keys = read_CMD(self.filename,default_keys=self.default_keys)  





#!/usr/bin/env python3
###########################################################################################
#  package:   RT
#  file:      __init__.py
#  brief:     init file
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#             Mladen Ivkovic 
#
# This file is part of pNbody.
###########################################################################################

from astropy import units as u
from astropy import constants as cte

from ..rtlib import * 

# some constants
kB = getkB()              # 1.3806488e-16  # erg/K
h_planck = geth_planck()  # 6.62606957e-27  # cm**2*g/s
c = getc()                # 29979245800.0  # cm/s
L_Sol = getL_Sol()        #3.827e33  # erg/s


# ionization levels
NuiHI   = getNuiHI()   #  3.288e15 Hz 
NuiHeI  = getNuiHeI()  #  5.945e15 Hz
NuiHeII = getNuiHeII() # 13.157e15 Hz
frequency_bins = [NuiHI,NuiHeI,NuiHeII] 


# ionization energies in eV
EiHI   = getEiHI()  *u.eV  #13.59807*u.eV
EiHeI  = getEiHeI() *u.eV  #24.58654*u.eV
EiHeII = getEiHeII()*u.eV  #54.41298*u.eV

# ionization frequencies in Hz
nuiHI  = (EiHI/cte.h).to(u.Hz)
nuiHeI = (EiHeI/cte.h).to(u.Hz)
nuiHeII= (EiHeII/cte.h).to(u.Hz)


def convert_photon_number_rate_to_luminosity_with_frequency_bins(T=1e5,Ndot=1e12,frequency_bins=frequency_bins):
  """
   -----------------------------------------------------------------------------
   Given a photon number and a minimal frequency, compute the corresponding
   luminosity in units of stellar luminosities assuming a blackbody spectrum
   for given frequency bins.
  
   Given the spectrum, we first compute the average ionizing photon energy for
   each frequency bin `i` by integrating the energy density and the number
   density from the lowest ionizing frequency to infinity (or rather until a
   very high maximal value):
  
     <E_photon>_i = [ \int_\nu J(\nu) d\nu ] / [ \int_\nu J(\nu) / (h \nu) d\nu ]
  
   with the integration boundaries [\nu_{i, min}, \nu_{i, max}] which are
   provided by the user.
  
   Then we need to compute the fraction `f` of ionizing photons in each bin:
  
    let  nl = \nu_{i,min},
         nu = \nu_{i,max}
  
     f_i = [ \int_{nl}^{nu}   J(\nu) / (h \nu) d\nu ] /
           [ \int_{nl}^\infty J(\nu) / (h \nu) d\nu ]
  
   Note that we start both integrations with `nl` because we assume that
   we need to distribute a certain amount of *ionizing* photons, not
   total photons following the given spectrum.
   We finally get the luminosity L from the photon number rate Ndot as
  
     L_i = <E_photon>_i * f_i * Ndot
  
   The final computation then simplifies to
  
     L_i = [ \int_{nl}^{nu}   J(\nu)           d\nu ] /
           [ \int_{nl}^\infty J(\nu) / (h \nu) d\nu ] * Ndot
           
   
   # temperature for blackbody spectrum
   T = 1e5  # K
   
   # ionizing photon number emission rate you want
   Ndot = 1e12  # s^-1 # for many Iliev tests

   # define upper limits for frequency bins. We assume that
   # the lowest bin is the first ionizing frequency.
   frequency_bins = [3.288e15, 5.945e15, 13.157e15]  # Hz
   
   # return
   luminosities in erg/s for each photon group
           
   -----------------------------------------------------------------------------
  """

  import scipy.integrate as integrate
  from .blackbody import B_nu, B_nu_over_h_nu, nu_peak


  peak_frequency = nu_peak(T, kB, h_planck)
  integration_limit = 100 * peak_frequency
  print("peak of the blackbody spectrum: {0:10.3e} [Hz]".format(peak_frequency))
  
  # Get the full integral over the photon number density
  number_density_integral_tot, nerr = integrate.quad(
      B_nu_over_h_nu, frequency_bins[0], integration_limit, args=(T, kB, h_planck, c)
  )
  
  Lsum = 0
  
  Luminosities_erg_per_s = []

  for f in range(len(frequency_bins)):
    # Set integration ranges.
    nu_min = frequency_bins[f]
    if f < len(frequency_bins) - 1:
        nu_max = frequency_bins[f + 1]
    else:
        # A potential problem is that the integral is non-convergent
        # for the methods used. So instead of integrating to infinity,
        # peak an upper limit in units of the peak frequency.
        nu_max = integration_limit

    print("Bin {0:3d}: {1:10.3e} - {2:10.3e} [Hz] ".format(f, nu_min, nu_max), end=" ")

    energy_density_integral, eerr = integrate.quad(
        B_nu, nu_min, nu_max, args=(T, kB, h_planck, c)
    )
    L = energy_density_integral / number_density_integral_tot * Ndot
    
    Luminosities_erg_per_s_cm2.append(L)

    print("Luminosity = {0:12.3e} [erg/s] {1:12.3e} [L_Sol]".format(L, L / L_Sol))

    Lsum += L

  # Sanity check: Sum of individual luminosities must correspond to luminosity
  # of a single frequency bin
  
  #  energy_density_integral, eerr = integrate.quad(
  #      B_nu, frequency_bins[0], integration_limit, args=(T, kB, h_planck, c)
  #  )
  #  number_density_integral, nerr = integrate.quad(
  #      B_nu_over_h_nu, frequency_bins[0], integration_limit, args=(T, kB, h_planck, c)
  #  )
  #
  #  L_single = energy_density_integral / number_density_integral * Ndot
  #
  #  print("Sanity check: sum over all bins / total luminosity integral = {0:12.3e}".format(Lsum / L_single))

  return Luminosities_erg_per_s_cm2












#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      libmiyamoto.py
#  brief:     Defines Miyamoto profiles
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np




def Vcirc(V, R, G=1.):
    """
    Mestel Circular velocity
    """
    return V




def SurfaceDensity(V, R, G=1.):
    """
    Mestel Density
    """

    return   V**2/(2*np.pi*G) /R








#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      nfw.py
#  brief:     Defines nfw profiles
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


"""
nfw model
"""

import numpy as np


class NFW():
  
  def __init__(self,rho0=None,rs=None,c=None,M200=None,H0=None,G=1.):
    """
    rho0 is the density at rs
    rs   is the specific radius
    c    is the concentration parameter
    M200 is the 'virial' mass
    rhoc is the critical density of the Universe
    """
    
    if (rho0 is not None) and (rs is not None):
      self.rho0 = rho0
      self.rs   = rs
      self.G    = G
      self.c    = None
    elif (c is not None) and (M200 is not None) and (H0 is not None): 
      rhoc = 3*H0**2/(8*np.pi*G)      
      r200 = np.cbrt(10*M200*G*H0)/(10*H0)
      dc = (200/3) * c**3/( np.log(1+c) - (c/(1+c)))       
      self.rho0 = dc*rhoc 
      self.rs   = r200/c
      self.G    = G
      self.c    = c
      
      self.rhoc = rhoc
      self.r200 = r200
      self.M200 = M200

  
  def info(self):
    print("NFW parameters")
    print("--------------")
    print("  c    = %g"%self.c)
    print("  rhoc = %g"%self.rhoc)
    print("  M200 = %g"%self.M200)
    print("  r200 = %g"%self.r200)
    print("  rs   = %g"%self.rs)
    print("  rho0 = %g"%self.rho0)
    
    
  def Potential(self,r):
    """
    return the Potential at a given radius r
    """
    return -4*np.pi*self.G*self.rho0*self.rs**2 * np.log(1+r/self.rs)/(r/self.rs)
    
  def CumulativeMass(self,r):
    """
    return the Cumulative mass at a given radius r
    """
    return   4*np.pi*self.rho0*self.rs**3 *   ( np.log(1+r/self.rs) - (r/self.rs)/(1+r/self.rs) )     
    
  def Density(self,r):
    """
    return the Density at a given radius r
    """
    return self.rho0 / ( (r/self.rs)**1 * (1+(r/self.rs))**2  )
  

  def Vcirc(self,r):
    """
    return the circular velocity at a given radius r
    """
    Mr = 4*np.pi*self.rho0*self.rs**3 * (   np.log(1+r/self.rs) - (r/self.rs)/(1+r/self.rs)       )
    return np.sqrt(self.G*Mr/r)
      
  
  def Kappa2(self,r):
      """
      return Kappa2 at a given radius r
      """
      kappa2 = -(self.rs/r**3)*np.log(1+r/self.rs) + (1/r**2)*(1./(1+r/self.rs)) - (1/(self.rs*r))*(1./(1+r/self.rs))**2
      kappa2 = -4*np.pi*self.G*self.rho0*self.rs**2  * kappa2
      return kappa2
  
  
  def Kappa(self, r):
      """
      return the Radial epicycle frequency at a given radius r
      """
      return np.sqrt(self.Kappa2(r))
  


##############################################
# The old implementation
##############################################



def Potential(rho0, a, r, G=1.):
    """
    Potential
    """
    return -4*np.pi*G*rho0*a**2 * np.log(1+r/a)/(r/a)

def CumulativeMass(rho0, a, r, G=1.):
    """
    Cumulative mass
    """
    return   4*np.pi*rho0*a**3 *   ( np.log(1+r/a) - (r/a)/(1+r/a) ) 
    

def Density(rho0, a, r, G=1.):
    """
    Density
    """
    return rho0 / ( (r/a)**1 * (1+(r/a))**2  )
    

def Vcirc(rho0, a, r, G=1.):
    """
    circular velocity
    """
        
    Mr = 4*np.pi*rho0*a**3 * (   np.log(1+r/a) - (r/a)/(1+r/a)       )
    return np.sqrt(G*Mr/r)
    

def Kappa2(rho0, a, r, G=1.):
    """
    Kappa2 
    """
    
    kappa2 = -(a/r**3)*np.log(1+r/a) + (1/r**2)*(1./(1+r/a)) - (1/(a*r))*(1./(1+r/a))**2
    kappa2 = -4*np.pi*G*rho0*a**2  * kappa2

    return kappa2


def Kappa(rho0, a, r, G=1.):
    """
    Radial epicycle frequency
    """
    return np.sqrt(Kappa2(rho0, a, r, G))



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:     integration_utils.py
#  brief:     Contains utilities fo coordinates system manipulation
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Darwin Roduit <darwin.roduit@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
from astropy import units as u
import astropy.coordinates as coord

coord.galactocentric_frame_defaults.set("v4.0")


#Data coming from Battaglia 2023 (http://arxiv.org/abs/2106.08819)
r_sun = 8.129*u.kpc
v_sun = np.array((11.1, 241.24, 7.25))*u.km/u.s


def compute_distance(distance_modulus):
    """Comute the distance with the distance modulus of an object."""
    return 10**(distance_modulus/5.0 + 1)


def convert_ICRS2carth(distance_modulus, right_ascension, declination,
                               pm_RA, pm_DEC, v_radial, unit_length,
                               unit_velocity):
    """Convert ICRS coordinates to galactocentric carthesian coordinates."""
    dist_particles = compute_distance(distance_modulus)*u.pc
    dist_particles = dist_particles.to(u.kpc).value
    ics = coord.SkyCoord(frame='icrs', ra=right_ascension*u.degree, dec=declination*u.degree,
                         distance=dist_particles*u.kpc,
                         pm_ra=pm_RA*u.mas/u.yr,
                         pm_dec=pm_DEC*u.mas/u.yr,
                         radial_velocity=v_radial*u.km/u.s,
                         differential_type=coord.SphericalDifferential)
    ics = ics.transform_to(coord.Galactocentric(galcen_distance=r_sun,
                                                galcen_v_sun=v_sun))
    # Using the parameters given in sec7, p. 13 from Battaglia et al.
    position = ics.data.get_xyz().to(unit_length)
    velocity = ics.velocity.get_d_xyz().to(unit_velocity)
    return position, velocity

def convert_sph2carth(r, phi, theta):
    """Converts spherical coordinates to carthesian coordinates."""
    x = r*np.cos(phi)*np.sin(theta)
    y = r*np.sin(phi)*np.sin(theta)
    z = r*np.cos(theta)
    return x, y, z


def convert_cyl2carth(R, phi, z):
    """Converts cylindrical coordinates to carthesian coordinates."""
    x = R*np.cos(phi)
    y = R*np.sin(phi)
    return x, y, z


def convert_cart2cyl(x, y, z):
    """Converts carthesian coordinates to cylindrical coordinates."""
    R = np.sqrt(x*x + y*y)
    # Determines phi
    if (x == 0) and (y == 0):
        phi = 0
    elif (x == 0) and (y != 0):
        phi = np.pi/2*y/np.abs(y)
    elif (x > 0):
        phi = np.arctan(y/x)
    elif (x < 0) and (y >= 0):
        phi = np.arctan(y/x)+np.pi
    elif (x < 0) and (y < 0):
        phi = np.arctan(y/x)-np.pi

    return R, phi, z


#!/usr/bin/python
###########################################################################################
#  package:   pNbody
#  file:      script_utils.py
#  brief:     Contains utilities classes and functions for scripts in the
#             scripts folder
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Darwin Roduit <darwin.roduit@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import argparse
import numpy as np


class ExplicitDefaultsHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
    """Prints the default arguments when they are provided."""

    def _get_help_string(self, action):
        if action.default in (None, False):
            return action.help
        return super()._get_help_string(action)


class RawTextArgumentDefaultsHelpFormatter(ExplicitDefaultsHelpFormatter,
                                           argparse.RawTextHelpFormatter):
    """Provides the defaults values and allows to format the text help."""
    pass


class store_as_array(argparse._StoreAction):
    """Provides numpy array as argparse arguments."""

    def __call__(self, parser, namespace, values, option_string=None):
        values = np.array(values)
        return super().__call__(parser, namespace, values, option_string)



###########################################################################################
#  package:   Mockimgs
#  file:      parameters.py
#  brief:     Parameters class
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


class Parameters():
  '''
  Define a parameters container, a slightly extended simple object.
  
  Note: part of this file is a copy/paste of 
  the argparse.Namespace class
  '''
  
  def __init__(self, **kwargs):
    for name in kwargs:
        setattr(self, name, kwargs[name])
  
  
  def __repr__(self):
      type_name = type(self).__name__
      arg_strings = []
      star_args = {}
      for arg in self._get_args():
          arg_strings.append(repr(arg))
      for name, value in self._get_kwargs():
          if name.isidentifier():
              arg_strings.append('%s=%r' % (name, value))
          else:
              star_args[name] = value
      if star_args:
          arg_strings.append('**%s' % repr(star_args))
      return '%s(%s)' % (type_name, ', '.join(arg_strings))

  def _get_kwargs(self):
      return list(self.__dict__.items())

  def _get_args(self):
      return []

  def __eq__(self, other):
      if not isinstance(other, Namespace):
          return NotImplemented
      return vars(self) == vars(other)

  def __contains__(self, key):
      return key in self.__dict__
        
  
  def update_from_dic(self,dic):
    '''
    update the parameter from a dictionary
    '''       
    for p in dic:
      if not hasattr(self,p):
        msg = "parameter %s is not a valid parameter"%p
        raise NameError(msg)
      else:
        setattr(self,p,dic[p])      


  def update_from_options(self,opt,skip_None=False):
    '''
    update the parameter from options
    '''      
    for key in opt.__dict__: 
      if self.__contains__(key):
        if skip_None and getattr(opt,key) is None: continue
        # update the value    
        setattr(self,key,getattr(opt,key))
        
        





###########################################################################################
#  package:   Mockimgs
#  file:      instrument.py
#  brief:     instrument class
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

from astropy import units as u
import numpy as np
import pickle
import datetime
import os
import socket
import pNbody
from . import filters
from . import lib


class Instrument():
  '''
  Define the Instrument class.
  An instrument is composed of a telescope, a detector (ccd) and a filter.
  '''
  def __init__(self,name=None,telescope=None,ccd=None,ifu=None,mos=None,filter_type=None,obs_object=None,mapping_method=None):
    
    self.name              = name
    self.telescope         = telescope
    
    self.ccd               = ccd
    if ccd is not None:
      self.ccd.instrument  = self
      
    self.ifu               = ifu
    if ifu is not None:
      self.ifu.instrument  = self  

    self.mos               = mos
    if mos is not None:
      self.mos.instrument  = self  
    
    self.filter            = filter_type
    
    self.object            = obs_object
    if obs_object is not None:
      self.object.instrument = self
        
    # send the telescope focal to the CCD is the latter is defined
    if self.ccd is not None:
      if self.telescope is not None:
        if self.telescope.focal is not None:
          self.ccd.set_focal(self.telescope.focal)
    
    # ccd mapping method  : spline or gauss or gauss_old
    if mapping_method is None:
        self.mapping_method = "gauss"


  def info(self):
    """
    get info on the instrument
    """
    print("name               : %s"%self.name)
    print("fov                : [%s,%s]"%(self.fov()[0].to(u.degree),self.fov()[1].to(u.degree)))
    self.telescope.info()
    self.ccd.info()
    self.filter.info()
    if self.object is not None:
      self.object.info()
      

  def fov(self):
    """
    get the instrument (ccd) field of view
    """    
    return self.ccd.fov
  
  
  def change_fov(self,fov):
    """
    Use only a subset of the available fov
    
    fov : new field of view
    """
    
    # define the field of view
    self.ccd.change_fov(fov)
    
  def get_arcsec_to_pixel_factors(self):
    """
    return the factor to convert arcsec to pixels
    """
    return self.ccd.get_arcsec_to_pixel_factors()

  def getFocal(self):
    """
    return the telescope focal if defined
    """
    if self.telescope is not None:
      return self.telescope.get_focal()
    else:
      raise ValueError("no telsecope defined. Cannot get the telescope focal.")
      
  def getObject(self):
    """
    return the object
    """
    if self.object is not None:
      return self.object
    else:
      raise ValueError("no object defined.")    


  def add_object(self,obj):
    """
    add an object to observe
    """
    self.object = obj
    self.ccd.set_extension_in_spatial_length(self.object.distance)
    
  def set_object_for_exposure(self,getAgeMH=False):
    """
    Open the N-body model and eventually compute missing quantities. 
    Rotate according to a given los.
    Scale according to a given distance.
    """
    self.object.open(getAgeMH=getAgeMH)    
    self.object.set_for_exposure()
    
    
    
  def set_filter(self,filtername):
    """
    set the filter
    
    filtername : name of the filter, e.g. : "BastI_GAIA_G"
                 of filter object  , e.g. : filters.Filter("BastI_GAIA_G")
        
    """
    
    if type(filtername) is str:
      self.filter = filters.Filter(filtername)
    else:
      self.filter = filtername 

  def set_mapping_method(self,mapping_method):
    """
    Set the kernel used to map particles to the CCD.

    mapping_method: 'gauss_old', 'gauss', 'spline'
    """
    self.mapping_method = mapping_method
    return

  def get_parameters(self,fmt=None):
    """
    return a dictionary containing usefull parameters
    """
    params = {}
    
    if fmt=="fits":
      params["NAME"]             = (self.name,"Instrument name")    
    else:  
      params["name"]             = self.name    

    # add telescope
    params.update(self.telescope.get_parameters(fmt)) 
    
    # add filter
    params.update(self.filter.get_parameters(fmt)) 
    
    # add object
    params.update(self.object.get_parameters(fmt))    
    
    # add ccd
    params.update(self.ccd.get_parameters(fmt))
    
    return params
    
  def ComputeFluxMap(self,magFromField=None):  
    """
    Compute a luminosity of flux map.
    Here the nbody model is supposed
    to be in the right units
    
    It return a 2D matrix containing either a fluxes or masses or luminosities
    
    If magFromField is not none, the magnitude is not computed but taken form the
    nbody object with the field name magFromField.
    """
    
    # get the nbody model
    nb = self.object
    
    # set map bins
    binsx = np.linspace(self.ccd.xmin,self.ccd.xmax,self.ccd.nx())
    binsy = np.linspace(self.ccd.ymin,self.ccd.ymax,self.ccd.ny())
    
    # get the magnitude per stellar particle

    if magFromField is None:
      # compute the magnitude for the defined filter:
      M = self.filter.Magnitudes(nb.mass,nb.age,nb.mh)     
    else:
      # take the magnitude from the hdf5 file
      M = getattr(nb,magFromField)
      
                
              
    # convert to flux (ignore the zero point)
    F = 10**(-M/2.5)
      
    # store in the mass field
    nb.mass = F.astype(np.float32) 
    
    
    # store the total mass
    total_mass = nb.mass.sum()

    # create the map
    if self.mapping_method is "gauss_old":
      raise ValueError("The mapping method gauss_old is not longer supported. Please, use spline or gauss instead.")
    
    elif self.mapping_method is "spline":
      FluxMap = self.ccd.MapUsingSpline(nb.pos,nb.mass,nb.rsp)
    
    elif self.mapping_method is "gauss":
      FluxMap = self.ccd.MapUsingGauss(nb.pos,nb.mass,nb.rsp)

  
    
    # compute total flux 
    total_flux = np.sum(np.ravel(FluxMap))
    
    # compute difference
    flux_diff = np.fabs(total_flux-total_mass)/total_mass
    #print("Flux difference : %g %%"%(flux_diff*100))
    
      
    # make it global
    self.FluxMap = FluxMap
    
    return self.FluxMap
  

  def FilterFluxMap(self,filter_opts={'mode':None}):
    """
    Filter the flux map
  
    FluxMas      : the flux map (numpy.array)
    filter_opts  : filter options
    """
    

    if hasattr(filter_opts,'psf_filename'):
      if filter_opts.psf_filename is not None:
        # redefine filter_opts
        opts = {}
        opts["mode"]              = "psf-convolution"
        opts["psf_filename"]      = filter_opts.psf_filename
        opts["image_pixel_scale"] = self.ccd.pixel_fov[0].to(u.arcsec).value        
        self.FluxMap = lib.FilterFluxMap(self.FluxMap,opts)
      
    
    return self.FluxMap
    

  def SurfaceBrightnessMap(self):
    """
    Convert a mass/luminosity/flux to a surface brightness map
    """
    
    # compute the map
    self.SBMap = lib.FluxToSurfaceBrightness(self.FluxMap,self.object.distance,self.ccd.pixel_area)
    
    return self.SBMap
    

  def getFluxMap(self):
    """
    return the current flux map
    """
    return self.FluxMap    
    
    
  def getSurfaceBrightnessMap(self):
    """
    return the current surface brightness map
    """
    return self.SBMap  


  def saveFitsImage(self,image,filename,units=None,comments=None,compress=True):
    """
    save a fits image
    
    image       : name of the image
    filename    : output file name
    units       : pixel units (str)
    comments    : additional comments (list)
    
    """
    from astropy.io import fits
    import os

    hdu = fits.PrimaryHDU(image)
    
    # add pixel unit
    hdu.header["UNITS"]     = (units,"pixel units")
    
    # add date
    hdu.header["TIME"]      = (str(datetime.datetime.now()),"creation time")
    
    # add user
    try:
      hdu.header["USER"]      = (os.getlogin(),"creator name")
    except:
      if "USER" in os.environ:
        hdu.header["USER"]      = (os.environ['USER'],"creator name")
      else:
        hdu.header["USER"]      = ('unknown',"creator name")  

    # add hostname
    try:
      hdu.header["HOSTNAME"]  = (socket.gethostname(),"hostname")
    except:
      hdu.header["HOSTNAME"]  = ('unknown',"hostname")
        
    # add user
    try:
      hdu.header["GITTAG"]    = (pNbody.__version__,"pNbody git tag")
    except:
      hdu.header["GITTAG"]    = ("unknown","pNbody git tag")
            
  
    #add other info 
    params = self.get_parameters(fmt="fits")
    for key in params:
      hdu.header[key] = params[key]
  
    # add comments
    if comments is not None:
      for comment in comments:
        hdu.header["COMMENT"]   = comment
    
    
    if os.path.isfile(filename):
      os.remove(filename)
    
    hdu.writeto(filename)    
    
    if compress:
      import gzip
      import shutil
      
      with open(filename, 'rb') as f_in:
        with gzip.open("%s.gz"%filename, 'wb') as f_out:
          shutil.copyfileobj(f_in, f_out)
      
      os.remove(filename)



  def draw(self,ax,mode=None,unit=None):
    """
    draw using matplotlib
    """    
    if self.ccd is not None:
      self.ccd.draw(ax,mode,unit)
      
    if self.ifu is not None:  
      self.ifu.draw(ax,mode,unit)

    if self.mos is not None:  
      self.mos.draw(ax,mode,unit)
    
    if self.object is not None:
      self.object.draw(ax,mode,unit)
    










#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      cosmo.py
#  brief:     Defines function related to cosmology
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


from . import ctes
# from . import units
from pNbody import cosmolib
# import types
import numpy as np

try:
    from scipy import optimize
    is_scipy = True
except ImportError:
    is_scipy = False


##########################################################################
#
# SOME COSMOLOGICAL RELATIONS
#
##########################################################################

UnitLength_in_cm = 3.085678e+21
UnitMass_in_g = 1.989e+43
UnitVelocity_in_cm_per_s = 100000.0
UnitTime_in_s = UnitLength_in_cm / UnitVelocity_in_cm_per_s
toGyrs = UnitTime_in_s / (60 * 60 * 24 * 365 * 1e9)


HUBBLE = 3.2407789e-18  # Hubble param in h/sec
HubbleParam = 0.73		# Hubble param in 100 km/s/Mpc
OmegaLambda = 0.76
Omega0 = 0.24

# Hubble constance in code unit
Hubble = HUBBLE * UnitTime_in_s

# deflauts cosmo parameters
defaultpars = {
    "Hubble": Hubble,
    "HubbleParam": HubbleParam,
    "OmegaLambda": OmegaLambda,
    "Omega0": Omega0}


###################
def setdefault():
    ###################
    """
    set default cosmological parameters
    """
    global defaultpars
    defaultpars = {
        "Hubble": Hubble,
        "HubbleParam": HubbleParam,
        "OmegaLambda": OmegaLambda,
        "Omega0": Omega0}


###################
def Z_a(a):
    ###################
    """
    z(a)
    """
    return 1. / a - 1

###################


def A_z(z):
    ###################
    """
    a(z)
    """
    return 1 / (z + 1)


###################
def Rho_c(localsystem_of_units):
    ###################
    """
    Critical density
    """

    G = ctes.GRAVITY.into(localsystem_of_units)
    H = ctes.HUBBLE.into(localsystem_of_units)

    rhoc = 3 * H**2 / (8 * np.pi * G)

    return rhoc


###################
def Hubble_a(a, pars=defaultpars):
    ###################
    """
    H(a)
    """
    OmegaLambda = pars['OmegaLambda']
    Omega0 = pars['Omega0']
    Hubble = pars['Hubble']

    hubble_a = Omega0 / (a * a * a) + (1 - Omega0 -
                                       OmegaLambda) / (a * a) + OmegaLambda
    hubble_a = Hubble * np.sqrt(hubble_a)

    return hubble_a

###################


def dt_da(da, a, pars=defaultpars):
    ###################
    """
    dt from da
    in units of 1/Hubble
    """

    dt = da / (a * Hubble_a(a, pars))
    # * toGyrs/HubbleParam if we want Gyrs, assuming Hubble=0.1

    return dt

###################


def Adot_a(a, pars=defaultpars):
    ###################
    """
    da/dt
    """
    OmegaLambda = pars['OmegaLambda']
    Omega0 = pars['Omega0']
    Hubble = pars['Hubble']

    hubble_a = Omega0 / (a * a * a) + (1 - Omega0 -
                                       OmegaLambda) / (a * a) + OmegaLambda
    hubble_a = Hubble * np.sqrt(hubble_a)
    adot_a = hubble_a * a

    return adot_a

###################


def Age_a(a, pars=defaultpars):
    ###################
    """
    cosmic age as a function of a
    Return a physical value (free of h) in Gyrs
    """
    OmegaLambda = pars['OmegaLambda']
    Omega0 = pars['Omega0']
    Hubble = pars['Hubble']
    HubbleParam = pars['HubbleParam']

    if isinstance(a, np.ndarray):
        a = a.astype(float)
    else:
        a = np.array([a], float)

    # here, Hubble is assumed to be in the default gadget units
    # and should thus allways be 0.1

    age_a = cosmolib.Age_a(
        a,
        Omega0,
        OmegaLambda,
        Hubble) * toGyrs / HubbleParam

    return age_a

###################


def CosmicTime_a(a, pars=defaultpars):
    ###################
    """
    cosmic time as a function of a in internal units,
    ie, (1/h)
    """
    OmegaLambda = pars['OmegaLambda']
    Omega0 = pars['Omega0']
    Hubble = pars['Hubble']
    
    scalar = False

    if isinstance(a, np.ndarray):
        a0 = np.zeros(len(a), float)
        a1 = a.astype(float)
    else:
        a0 = np.array([0], float)
        a1 = np.array([a], float)
        scalar = True
        

    t0 = cosmolib.Age_a(a0, Omega0, OmegaLambda, Hubble)
    ta = cosmolib.Age_a(a1, Omega0, OmegaLambda, Hubble)
    
    if scalar:
      t0 = t0[0]
      ta = ta[0]
    
    return t0 - ta


###################
def a_CosmicTime(t, pars=defaultpars, a0=0.5):
    ###################
    """
    return a for a given cosmic time
    """

    if not is_scipy:
        raise Exception("module scipy needed for function a_CosmicTime !")

    def zfct(a, t, pars):
        t = CosmicTime_a(a, pars)[0] - t
        return t

    a = optimize.newton(zfct, x0=a0, args=(t, pars), tol=1e-5)
    return a



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      __init__.py
#  brief:     
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
import sys, os
import time

from . import BastI
from . import CMD
from . import MIST

database_directory=None
default_keys=None


#'''
#
## setup database directory
#if "PNBODY_ISOCHRONE_DEFAULT_DATABASE" in os.environ:
#  database_directory = os.environ["PNBODY_ISOCHRONE_DEFAULT_DATABASE"]
#else:
#  raise NameError("Environnement variable %s not defined ! Please define it."%("PNBODY_ISOCHRONE_DEFAULT_DATABASE"))
#
## setup database directory
#if "PNBODY_ISOCHRONE_LIBRARY_NAME" in os.environ:
#  filter_system = os.environ["PNBODY_ISOCHRONE_LIBRARY_NAME"]
#else:
#  raise NameError("Environnement variable %s not defined ! Please define it."%("PNBODY_ISOCHRONE_LIBRARY_NAME"))
#
#'''
#
## default keys of the database
##default_keys = ["log10_isochrone_age_yr","initial_mass","log_Teff","log_L","[Fe/H]_init","SDSS_u","SDSS_g","SDSS_r","SDSS_i","SDSS_z"]
#default_keys = None
#
##if not os.path.isdir(database_directory):
##   raise NameError("directory %s not found !"%(database_directory))
# 
#database_directory=None
#def getDatabaseDirectory():
#  return database_directory
#


def Isochrones(dirname=database_directory,default_keys=default_keys,filename=None):
  
  if filename is not None:
    if os.path.isfile(filename):
      return CMD.Grid(filename=filename,default_keys=default_keys)
    else:
      raise NameError("The file %s does not exists !"%filename)
  
  else:
    
    if os.path.isdir(dirname):
      if os.path.basename(dirname)[:4] == "MIST":
        return MIST.Grid(dirname=dirname,default_keys=default_keys)
      else:
        return BastI.Grid(dirname=dirname,default_keys=default_keys)
    
    else:
      raise NameError("The directoy %s does not exists !"%dirname)
    
    






